import os.path
import pickle
from enum import Enum
from typing import Iterable, Literal, Union, Any

import numpy as np
import pyvista as pv
import tqdm

from .algos import QuickUnion, ConnectedTriangleSurfaces
from .dhash import _hash_array
from .file import make_cache_file, make_cache_directory
from .hash import hash_numpy_array
from .obj_splitter import ObjSplitter
from .string import parse_axes_labels
from .time import Timer


DataSetLike = Union[pv.DataSet, pv.MultiBlock]
DataSetLikeType = (pv.DataSet, pv.MultiBlock)


def hash_model(model, n_samples=10):
    return hash_numpy_array(model.points, n_samples)


def dhash_model(model, n_samples=10):
    return _hash_array(model.points, n_samples)


def as_pyvista_array(arr):
    arr = np.asarray(arr)
    if arr.dtype == bool:
        arr = arr.astype(np.int8)

    return arr


def extract_model_list_bounds(model_list: list[pv.DataSet]):
    bounds = None
    for m in model_list:
        if bounds is None:
            bounds = np.asarray(m.bounds)
            continue
        bounds[1::2] = np.max([bounds[1::2], m.bounds[1::2]], axis=0)
        bounds[0::2] = np.min([bounds[0::2], m.bounds[0::2]], axis=0)

    return bounds


def merge_as_multiblock(model_list):
    mb = pv.MultiBlock()
    for m in model_list:
        mb.append(m)

    return mb


def load_model_file(model_file, verbose=True):
    model_name = os.path.basename(model_file)
    cache_path = make_cache_file(model_name + '.vtk')

    if os.path.exists(cache_path):
        if verbose:
            print(f'Loading cached model from {cache_path}...')
        model_file = cache_path

    reader = pv.get_reader(model_file)

    if verbose:
        reader.show_progress()

    timer = Timer()
    with timer:
        model = reader.read()

    if timer.elapsed > 5 and cache_path != model_file:
        if verbose:
            print(f'Saving loaded model to {cache_path}...')
        model.save(cache_path, binary=True)

    return model


def load_grouped_file(model_file, verbose=True):
    if os.path.splitext(model_file)[1] != '.obj':
        return None  # 目前只支持obj文件

    splitter = ObjSplitter(model_file)
    model_name = os.path.basename(model_file)
    cache_path = make_cache_directory(os.path.splitext(model_name))
    if not os.path.exists(cache_path):
        splitter.split_by_group(cache_path)

    models = []
    for sub_model_file in os.listdir(cache_path):
        sub_model_path = os.path.join(cache_path, sub_model_file)
        models.append(load_model_file(sub_model_path))

    return models


def split_models_in_memory(model, verbose=True):
    cache = make_cache_file(f'model_division/{dhash_model(model)}.sub')
    if os.path.exists(cache):
        with open(cache, 'rb') as f:
            if verbose:
                print(f'Loading split models from {cache}...')
            ret = pickle.load(f)
    else:
        models = split_models_in_memory_pointwisely(model, verbose=verbose)
        ret = []

        if verbose:
            progress = tqdm.tqdm(total=model.n_faces, desc='Splitting models by edge connectivity')
        else:
            progress = None

        for m in models:
            ms = split_models_in_memory_edgewisely(m, verbose=verbose, progress=progress)
            ret.extend(ms)

        with open(cache, 'wb') as f:
            if verbose:
                print(f'Saving split models to {cache}...')
            pickle.dump(models, f)

    return ret


def split_models_in_memory_edgewisely(model, verbose=True, progress=None):
    points = ConnectedTriangleSurfaces()
    model_faces = model.faces
    i = 0
    faces_size = len(model_faces)

    if verbose and progress is None:
        progress = tqdm.tqdm(total=faces_size, desc='Splitting models by edge connectivity')

    while i < faces_size:
        nv = model_faces[i]
        pts = model_faces[i + 1:i + nv + 1]
        i = i + nv + 1
        points.add(pts)
        if progress is not None:
            progress.update(1)

    models = []
    for g in points.get_groups():
        indices = np.asarray(list(g))
        sub_model = model.extract_points(indices, adjacent_cells=False)
        sub_model.clear_data()
        models.append(sub_model.extract_surface())

    return models


def split_models_in_memory_pointwisely(model, verbose=True):
    """
    按点连通性来划分子几何体，对一些情况无法适用
    """
    unions = QuickUnion(model.points.shape[0])
    model_faces = model.faces
    i = 0
    faces_size = len(model_faces)

    progress = None
    if verbose:
        progress = tqdm.tqdm(total=faces_size, desc='Splitting models by point connectivity')

    while i < faces_size:
        nv = model_faces[i]
        pts = model_faces[i + 1:i + nv + 1]
        i = i + nv + 1
        p1 = pts[0]
        for p2 in pts[1:]:
            unions.connect(p1, p2)

        if progress is not None:
            progress.update(nv + 1)

    if verbose:
        progress.close()

    unions.collapse(verbose)
    groups = np.unique(unions.unions)

    models = []

    if verbose:
        groups = tqdm.tqdm(groups, desc='Extracting sub-models')

    for g in groups:
        indices = np.argwhere(unions.unions == g).squeeze()
        sub_model = model.extract_points(indices, adjacent_cells=False)
        sub_model.clear_data()
        models.append(sub_model.extract_surface())

    return models


class ModelTranslation:
    def __init__(self, offset, inplace=False):
        self.offset = offset
        self.inplace = inplace

    def __call__(self, mesh):
        return pv_ufunc_apply(mesh, self._translate, inplace=self.inplace)

    def _translate(self, x):
        return x.translate(self.offset, inplace=True)


def pv_ufunc_apply(obj: DataSetLike, fn, *, inplace=True):
    """对PyVista对象执行指定函数。

    如果目标为 :class:`pyvista.DataSet` 及其子类，则直接将fn作用于obj；
    如果目标为 :class:`pyvista.MultiBlock` 则遍历其所有成员递归执行pv_ufunc_apply。

    Parameters
    ----------
    obj
        需要应用函数 `fn` 的PyVista对象
    fn
        需要应用到 `obj` 对象上的处理函数。
        语义上要求为 `inplace` 操作，具体 `inplace` 行为由 `inplace` 参数控制
    inplace
        指示操作是否直接在目标对象上进行。
        若为 `False` ，则会在所有对象的拷贝上应用 `fn` ，然后重新拼装完整对象。

    Returns
    -------
    obj
        obj自身，如果 `inplace` 为False，则为拷贝的新对象
    """
    if isinstance(obj, pv.MultiBlock):
        if inplace:
            for k in obj.keys():
                pv_ufunc_apply(obj[k], fn, inplace=inplace)
            return obj
        else:
            return pv.MultiBlock(
                {k: pv_ufunc_apply(obj[k], fn, inplace=inplace) for k in obj.keys()}
            )
    else:
        if not inplace:
            obj = obj.copy()
        fn(obj)
        return obj


def pv_ufunc_visit(obj: DataSetLike, fn):
    """遍历PyVista对象并执行指定函数，不会进行拷贝，如果 `fn` 产生了修改，则会直接修改 `obj`。

     即 `pv_ufunc_apply` 的强制 `inplace` 版本，语义上要求 `fn` 不会对DataSet进行修改，只进行遍历获取信息。

    Parameters
    ----------
    obj
        需要应用函数 `fn` 的PyVista对象
    fn
        需要应用到 `obj` 对象上的处理函数
    """
    pv_ufunc_apply(obj, fn, inplace=True)


def pv_ufunc_map(obj: DataSetLike, fn):
    """遍历PyVista对象并执行映射

    Parameters
    ----------
    obj
        需要应用函数 `fn` 的PyVista对象
    fn
        需要应用到 `obj` 对象上的映射
    """
    mapped = []

    def wrapper(dataset):
        mapped.append(fn(dataset))

    pv_ufunc_visit(obj, wrapper)

    return mapped


class DataAssociation(Enum):
    Cell = 'cell'
    Point = 'point'
    Row = 'row'

    Field = 'field'

    def __str__(self):
        return self.value

    @property
    def field_count(self):
        if self == DataAssociation.Field:
            return None
        else:
            return f'n_{self}s'

    @property
    def field_collection(self):
        return f'{self}_data'

    @property
    def field(self):
        return f'{self}s'

    def get(self, dataset):
        return getattr(dataset, self.field)

    def set(self, dataset, info):
        setattr(dataset, self.field, info)

    def get_count(self, dataset):
        return getattr(dataset, self.field_count)

    def get_collection(self, dataset):
        return getattr(dataset, self.field_collection)

    def get_data(self, dataset, key):
        return self.get_collection(dataset)[key]

    def set_data(self, dataset, key, val):
        self.get_collection(dataset).__setitem__(key, val)


def pv_ufunc_assign(obj: pv.DataSet,
                    data_type: DataAssociation,
                    key, val,
                    inplace=True,
                    set_active=True):
    """赋值给对象的数据字段，对MultiBlock则作用于其中的每个对象

    Parameters
    ----------
    obj
        待赋值对象
    data_type
        待赋值数据类型，例如 "`cell`", "`point`", "`field`"
    key
        待赋值字段类型
    val
        值，如果为单个值，则会通过np.full展开到对应字段长度
    inplace
        指示操作是否直接在目标对象上进行
    set_active
        指示是否将新赋予的值设置为默认向量

    Returns
    -------
    obj
        obj自身，如果`inplace`为False，则为拷贝的新对象
    """
    if inplace is False:
        obj = obj.copy()

    val = as_pyvista_array(val)

    def ufunc(x):
        element_count = data_type.get_count(x)

        if val.ndim == 0:
            _val = np.full(element_count, val)
        else:
            _val = val

        data_type.set_data(x, key, _val)

        if set_active:
            x.set_active_scalars(key)

        return x

    pv_ufunc_apply(obj, ufunc, inplace=inplace)


AxesLabels = Iterable[Union[Literal['x', 'y', 'z'], int]]


def reset_model_axes(model, new_axes: AxesLabels, inplace=False):
    new_axes = parse_axes_labels(new_axes, max_length=3)
    return reset_model_axes_unsafe(model=model, new_axes=new_axes, inplace=inplace)


def reset_model_axes_unsafe(model, new_axes: list[int], inplace=False):
    if not inplace:
        model = model.copy(deep=True)

    original_axes = []
    sub_axes = []
    for i, a in enumerate(new_axes):
        if i != a:
            original_axes.append(i)
            sub_axes.append(a)

    model.points[:, sub_axes] = model.points[:, original_axes]

    return model


def convert_model_dict_to_multiblock(model_dict: dict[Any, pv.DataSet], root=None):
    """将模型字典转换为pv.MultiBlock实例，支持嵌套字典

    Parameters
    ----------
    model_dict
        模型字典
    root
        基础MultiBlock实例

    Returns
    -------
    multi_block
        从模型字典构造的嵌套MultiBlock
    """
    if root is None:
        root = pv.MultiBlock()

    for k, v in model_dict.items():
        if isinstance(v, dict):
            v = convert_model_dict_to_multiblock(v)
        root[k] = v

    return root



import time


class Timer:
    def __init__(self):
        """单位为ns的计时器
        """
        self.started = 0
        self.elapsed = 0
        self.stopped = -1

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.check()

    def start(self):
        self.started = time.perf_counter_ns()

    def check(self):
        self.stopped = time.perf_counter_ns()
        self.elapsed = self.stopped - self.started

    def __str__(self):
        if self.stopped == -1:
            self.check()

        elapsed_mins, elapsed_secs, elapsed_milis, elapsed_micros = \
            perf_ns_epoch_time(self.started, self.stopped)
        if elapsed_mins > 0:
            return f"{elapsed_mins} min {elapsed_secs} s"
        elif elapsed_secs > 0:
            return f"{elapsed_secs} s {elapsed_milis} ms"
        elif elapsed_milis > 0:
            return f"{elapsed_milis + elapsed_micros / 1000:.5f} ms"
        else:
            return f"{elapsed_micros:.5f} us"


def epoch_time(start_time, end_time):
    elapsed_time = end_time - start_time
    elapsed_mins = int(elapsed_time / 60)
    elapsed_secs = int(elapsed_time - (elapsed_mins * 60))
    return elapsed_mins, elapsed_secs


def perf_ns_epoch_time(start_time, end_time):
    elapsed_time = end_time - start_time
    elapsed_micros = elapsed_time / 1000
    elapsed_milis = int(elapsed_micros / 1000)
    elapsed_secs = int(elapsed_milis / 1000)
    elapsed_mins = int(elapsed_secs / 60)

    elapsed_micros -= elapsed_milis * 1000
    elapsed_milis -= elapsed_secs * 1000
    elapsed_secs -= elapsed_mins * 60

    return elapsed_mins, elapsed_secs, elapsed_milis, elapsed_micros





def scipy2torch(src, device='cpu'):
    """Convert scipy.sparse.csr_matrix to torch.sparse_csr_tensor.
    ref: https://github.com/pearu/pearu.github.io/blob/1cdb9eaf60e1313834d7dedb80feb74c1153f3fa/csr_naming_conventions.md
    """
    import torch
    import scipy.sparse

    if isinstance(src, scipy.sparse.dia_matrix):
        src = src.tocoo()  # TODO: 直接转换到torch.sparse

    if isinstance(src, scipy.sparse.csr_matrix):
        return torch.sparse_csr_tensor(
            torch.from_numpy(src.indptr),
            torch.from_numpy(src.indices),
            torch.from_numpy(src.data),
            device=device,)

    elif isinstance(src, scipy.sparse.csc_matrix):
        return torch.sparse_csc_tensor(
            torch.from_numpy(src.indptr),
            torch.from_numpy(src.indices),
            torch.from_numpy(src.data),
            device=device,)




import abc
import queue
from concurrent.futures import Future
from concurrent import futures
from functools import partial

from .executor import Executor
from .queue_like_worker_context import QueueLikeWorkerContext
from .utils import exception_caught, traverse_args


class PoolExecutor(Executor):
    def __init__(self, pool_executor: futures.Executor, workers):
        """构造基于Python标准Executor的执行器

        Parameters
        ----------
        pool_executor
            符合Python约定的工作单元池执行器
        """
        super().__init__()
        self.pool = pool_executor
        self.workers = workers

        self._queue = None

    @abc.abstractmethod
    def _get_queue(self):
        pass

    @property
    def queue(self):
        if self._queue is None:
            self._queue = self._get_queue()
        return self._queue

    @property
    def has_queue(self):
        return self._queue is not None

    def map(self, func, *iterables, worker=None, workers=None, chunksize=None):
        with self.monitor_events():
            return list(self.pool.map(func, *iterables, chunksize=chunksize))

    def do_submit(self, func, *args, workers=None, **kwargs):
        # 用于在func抛出异常时，打印异常栈信息然后传播
        # TODO: 让异常栈停留在异常抛出的位置，可能需要一些hack
        func = partial(exception_caught, func)

        # 自动将传入的future提取为值
        args, kwargs = traverse_args(args, kwargs,
                                     lambda x: x if not isinstance(x, Future) else self.gather([x])[0])
        return self.pool.submit(func, *args, **kwargs)

    def get_workers(self):
        return self.workers

    def is_local(self):
        return True

    def _gather_single(self, future):
        return future.result()

    def get_worker_context(self):
        return QueueLikeWorkerContext(self.queue)

    def receive_events(self, timeout):
        while True:
            try:
                event, msg = self.queue.get(True, timeout)
                yield event, msg
            except queue.Empty:
                yield None, None

    def check_if_events_thread_are_required(self):
        return (self.has_queue
                and super().check_if_events_thread_are_required())

    def shutdown(self, wait=True):
        self.pool.shutdown(wait=wait)



from metalpy.mexin.utils.type_map import assert_all_types_exists
from metalpy.scab.builder import SimulationBuilder


def test_builders_map():
    assert_all_types_exists(SimulationBuilder._registry)



from __future__ import annotations

import contextlib
import functools
from threading import Lock
from typing import TYPE_CHECKING, Optional, Sequence, TypeVar, Generator, overload

import tqdm


if TYPE_CHECKING:
    from metalpy.mepa import Executor, WorkerContext


_ElemT = TypeVar('_ElemT')


def _locally_locked(func: _ElemT) -> _ElemT:
    @functools.wraps(func)
    def wrapper(self: ParallelProgress, *args, **kwargs):
        if self._local_lock is None:
            return func(self, *args, **kwargs)
        else:
            with self._local_lock:
                return func(self, *args, **kwargs)

    return wrapper


class ParallelProgress:
    def __init__(self, executor: Optional['Executor'], *, total=0, **kwargs):
        """适用于 `mepa.Executor` 的并行进度条，
        支持直接传递给worker，以类似 `tqdm` 的形式并发更新进度

        如果初始化时指定了 `total` ，则创建者也有义务执行 `close` 进行关闭。

        Parameters
        ----------
        executor
            执行器
        total
            任务总量，也可以在各个worker中用 `register` 登记工作量
        kwargs
            其它要传给 `tqdm` 的关键字参数

        Notes
        -----
        Executor本地的所有事件响应是在单线程内进行的，
        因此如果对象通过WorkerContext通信更新进度的话，是不需要加锁的，
        需要创建时手动指定 `enable_local_lock(False)`

        但是例如ThreadExecutor这种本地的执行器，
        可以绕过WorkerContext通信更新进度，
        此时则需要对操作进行加锁，防止进度条数据错误

        Examples
        --------
        基本函数包含 `register` 、 `register` 和 `close` ，分别负责注册进度，更新进度和汇报进度完成。
        其中 `register` 和 `close` 应保证成对出现：

        >>> with ProcessExecutor(n_tasks) as executor:
        >>>     progress = executor.progress()
        >>>     def task():
        >>>         progress.register(10)  # 告知主机工作量
        >>>         for i in range(10):
        >>>             sleep(0.5)
        >>>             progress.update(1)  # 告知主机进度更新
        >>>         progress.close()  # 告知主机工作已完成（每个register应当与close配对）
        >>>     tasks = [executor.submit(task) for _ in range(n_tasks)]
        >>>     executor.gather(tasks)

        可以通过 `progress` 提供的上下文管理器来避免手动 `close` ：

        >>> with ProcessExecutor(4) as executor:
        >>>     progress = executor.progress()
        >>>     def task():
        >>>         with progress(10):
        >>>             for i in range(10):
        >>>                 sleep(0.5)
        >>>                 progress.update(1)  # 告知主机进度更新
        >>>     tasks = [executor.submit(task) for _ in range(4)]
        >>>     executor.gather(tasks)

        `progress` 也提供了类似 `tqdm` 的便捷函数 `range` 和 `iter` （ `trange` 和 `tqdm` ）：

        >>> with ProcessExecutor(4) as executor:
        >>>     progress = executor.progress()
        >>>     def task():
        >>>         for _ in progress.range(10):
        >>>             sleep(0.5)
        >>>     tasks = [executor.submit(task) for _ in range(4)]
        >>>     executor.gather(tasks)
        """
        self.executor = executor
        self.progress = None

        self.progress_configs = {
            'total': total,
            **kwargs
        }
        self.progress: tqdm.tqdm | None = None

        self.context: WorkerContext | None = None
        self._register_key = None
        self._update_key = None
        self._close_key = None

        if executor is not None:
            self.context = executor.get_worker_context()
            self._register_key = executor.register(self.register).event
            self._update_key = executor.register(self.update).event
            self._close_key = executor.register(self.close).event
            self.progress = tqdm.tqdm(**self.progress_configs)

        self.n_distributes = 0
        if total > 0:
            # 创建者制定了总量，也视为一次分发，需要由创建者来 `close`
            self.n_distributes += 1

        self._disable = False
        self._local_lock = Lock()

    @contextlib.contextmanager
    def __call__(self, total):
        self.register(total)
        try:
            yield
        finally:
            self.close()

    @property
    def disable(self):
        return self._disable

    def iters(self, iterable: Sequence[_ElemT]) -> Generator[_ElemT, None, None]:
        total = len(iterable)
        with self(total):
            for elem in iterable:
                yield elem
                self.update(1)

    @overload
    def range(self, end): ...

    @overload
    def range(self, start, end): ...

    @overload
    def range(self, start, end, step): ...

    def range(self, *args):
        yield from self.iters(range(*args))

    @_locally_locked
    def register(self, n):
        if self.executor is not None:
            self.n_distributes += 1
            self.progress.total += n
            self.progress.refresh()
        else:
            self.context.fire(self._register_key, n)

    @_locally_locked
    def update(self, n):
        if self.executor is not None:
            self.progress.update(n)
        else:
            self.context.fire(self._update_key, n)

    @_locally_locked
    def close(self):
        if self.executor is not None:
            if self.n_distributes > 0:
                self.n_distributes -= 1
                if self.n_distributes <= 0 and self.progress is not None:
                    self.dispose()
        else:
            self.context.fire(self._close_key)

        self._disable = True

    tqdm = iters
    trange = range
    reset = register

    def enable_local_lock(self, enable):
        """指定是否对进度条操作操作启用锁

        Parameters
        ----------
        enable
            指示是否启用
        """
        if enable:
            if self._local_lock is None:
                self._local_lock = Lock()
        else:
            self._local_lock = None

    def dispose(self):
        if self.executor is not None and self.context is not None:
            self.progress.close()
            self._register_key = self.executor.off(self._register_key)
            self._update_key = self.executor.off(self._update_key)
            self._close_key = self.executor.off(self._close_key)
            self.context = None

    def distribute(self):
        self.enable_local_lock(False)

        return {
            'executor': None,
            'context': self.context,
            '_disable': self._disable,
            '_register_key': self._register_key,
            '_update_key': self._update_key,
            '_close_key': self._close_key,
            '_local_lock': None
        }

    def __getstate__(self):
        return self.distribute()



import numpy as np
from SimPEG.potential_fields import magnetics
from discretize.utils import mkvc

from metalpy.scab import Tied, Demaged
from metalpy.scab.builder import SimulationBuilder
from metalpy.scab.demag.demagnetization import Demagnetization
from metalpy.scab.demag.utils import get_prolate_spheroid_demag_factor
from metalpy.scab.modelling.shapes import Ellipsoid
from metalpy.utils.file import make_cache_directory
from metalpy.utils.numeric import limit_significand


def test_numerical_solvers():
    # 基于各种求解器求解退磁效应
    pred = {}
    for method in Demagnetization.methods:
        pred[method] = predict_with_demag(method=method)

    ref = Demagnetization.Compressed
    tmi_ref, model_ref = pred.pop(ref)
    for method, (tmi_pred, model_pred) in pred.items():
        # 测试不同方法结果是否有偏差
        a_and_b = f'`{ref}` and `{method}`'
        np.testing.assert_allclose(
            tmi_pred, tmi_ref, err_msg=f'TMI MAPE between {a_and_b}.',
            rtol=1e-3
        )
        np.testing.assert_allclose(
            model_pred, model_ref, err_msg=f'Demag model MAPE between {a_and_b}.',
            rtol=1e-1
        )


def test_numerical_solution():
    # 基于椭球体退磁因子计算
    tmi_truth, model_truth = predict_with_demag(by_factor=True)

    # 基于积分方程法求解退磁效应
    tmi_pred, model_pred = predict_with_demag(method=Demagnetization.Seperated)

    # 测试压缩求解器的结果是否和解析解有偏差
    tmi_mape = np.mean(np.abs((tmi_truth - tmi_pred) / tmi_truth))
    assert tmi_mape < 0.2

    model_mape = np.mean(np.abs((model_truth - model_pred) / model_truth))
    assert model_mape < 0.35


def predict_with_demag(
        by_factor=False,
        method=None,
        cell_size=(2, 2, 2)
):
    cell_size = limit_significand(cell_size)
    shape, factor = make_test_shape_and_demag_factor()
    model_mesh = shape.to_scene(model=80).build(cell_size=cell_size)

    if by_factor:
        demag = Demaged(factor=factor)
    else:
        demag = Demaged(method=method)

    builder = SimulationBuilder.of(magnetics.simulation.Simulation3DIntegral)
    builder.patched(Tied(arch='cpu'), demag)
    builder.source_field(50000, 45, 20)
    builder.receivers(make_test_pts())
    builder.active_mesh(model_mesh)
    builder.model_type(scalar=True)
    builder.store_sensitivities(make_cache_directory(f'demag_sens/{cell_size}'))
    simulation = builder.build()
    tmi_pred = simulation.dpred(model_mesh.model)
    model_pred = simulation.chi

    return tmi_pred, model_pred


def make_test_shape_and_demag_factor():
    a, c = 10, 40
    shape = Ellipsoid.spheroid(a, c, polar_axis=0)
    factor = get_prolate_spheroid_demag_factor(c / a, polar_axis=0)

    return shape, factor


def make_test_pts():
    obsx = np.linspace(-512, 512, 64 + 1)
    obsy = np.linspace(-512, 512, 64 + 1)
    obsx, obsy = np.meshgrid(obsx, obsy)
    obsx, obsy = mkvc(obsx), mkvc(obsy)
    obsz = np.full_like(obsy, 80)
    receiver_points = np.c_[obsx, obsy, obsz]

    return receiver_points



from __future__ import annotations

from SimPEG import maps
from SimPEG.simulation import LinearSimulation

from . import SimulationBuilder


class LinearSimulationBuilder(SimulationBuilder):
    def __init__(self, sim_cls: LinearSimulation):
        super().__init__(sim_cls)

    @SimulationBuilder._supplies('model_map')
    def model_map(self, mapping: maps.IdentityMap):
        return mapping



from __future__ import annotations

from typing import Literal, Iterable

import numpy as np
from numpy.linalg import norm


class ClippablePolygon:
    def __init__(self, pts):
        self.winding = determine_winding(pts)
        self.pts = pts
        self.nexts = np.arange(1, pts.shape[0] + 1, dtype=np.int64)
        self.nexts[-1] = 0
        self.prevs = np.arange(-1, pts.shape[0] - 1, dtype=np.int64)
        self.prevs[0] = pts.shape[0] - 1

    def remove(self, index):
        n = self.next(index)
        p = self.prev(index)
        self.prevs[n] = p
        self.nexts[p] = n

    def regulate_index(self, index):
        if index < 0:
            index += self.pts.shape[0]
        elif index >= self.pts.shape[0]:
            index -= self.pts.shape[0]
        return index

    def prev(self, index):
        index = self.regulate_index(index)
        return self.prevs[index]

    def next(self, index):
        index = self.regulate_index(index)
        return self.nexts[index]

    def extract_triangle(self, index):
        indices = [self.prev(index), index, self.next(index)]
        return self.pts[indices]

    def check_index(self, index):
        return self.prev(self.next(index)) == index


class VertexTriangle:
    def __init__(self, index, polygon: ClippablePolygon):
        self.index = index
        self.polygon = polygon
        self.triangle = None
        self.angle = 0
        self.conflicted_vertex: VertexTriangle | Literal[False] | None = None

        self.prev = self
        self.next = self

        self.assemble_triangle()

    def append(self, node: 'VertexTriangle'):
        last = self.prev
        last.next = node
        node.prev = last

        self.prev = node
        node.next = self

    def assemble_triangle(self):
        self.triangle = self.polygon.extract_triangle(self.index)
        a = self.triangle[1] - self.triangle[0]
        b = self.triangle[2] - self.triangle[1]
        self.angle = cross(a, b) / norm(a) / norm(b)

        self.invalidate()

    @property
    def reflex(self):
        # 判断是否为内凹顶点
        return self.polygon.winding * self.angle < 0

    @property
    def vertex(self):
        return self.triangle[1]

    @property
    def prev_index(self):
        return self.polygon.prev(self.index)

    @property
    def next_index(self):
        return self.polygon.next(self.index)

    def clip(self):
        self.polygon.remove(self.index)
        return self.prev.index, self.index, self.next.index

    def remove(self):
        self.prev.next = self.next
        self.next.prev = self.prev

    def invalidate(self):
        self.conflicted_vertex = None

    def test_is_ear_many(self, vertices: Iterable['VertexTriangle']):
        if self.conflicted_vertex in vertices:
            return False

        for v in vertices:
            if not self.test_inside(v):
                return False

        return True

    def test_inside(self, vertex: 'VertexTriangle'):
        if vertex in [self, self.prev, self.next]:
            return True  # 顶点不认为在三角形内
        if is_inside_convex_polygon(vertex.vertex, self.triangle):
            self.conflicted_vertex = vertex
            return False
        else:
            return True

    def __hash__(self):
        return hash(self.index)

    def __iter__(self):
        ptr = self
        yield ptr
        ptr = ptr.next
        while ptr != self:
            yield ptr
            ptr = ptr.next


def segment_area_times_2(x2, x1, y2, y1):
    return (x2 - x1) * (y2 + y1)


def determine_winding(pts):
    """计算给定有序二维点集的时钟序

    Parameters
    ----------
    pts
        有序二维点集

    Returns
    -------
    ret
        -1代表顺时针方向，1代表逆时针方向
    """
    v = np.sum(segment_area_times_2(pts[1:, 0], pts[:-1, 0], pts[1:, 1], pts[:-1, 1]))
    v += segment_area_times_2(pts[0, 0], pts[-1, 0], pts[0, 1], pts[-1, 1])
    return -np.sign(v)


def cross(la, lb):
    return la[0] * lb[1] - la[1] * lb[0]


def determine_line_winding(la, lb) -> int:
    return np.sign(cross(la, lb))


def is_inside_convex_polygon(p, pts) -> bool:
    """判断给定点p是否存在于凸多边形中

    Parameters
    ----------
    p
        给定点p
    pts
        凸多边形顶点

    Returns
    -------
    ret
        p是否存在于凸多边形中
    """
    vectors = pts - p
    winding = determine_line_winding(vectors[-1], vectors[0])
    if winding == 0:
        return True
    for va, vb in zip(vectors[:-1], vectors[1:]):
        w = determine_line_winding(va, vb)
        if w != winding:
            return False

    return True


def ear_clip(pts, verbose=True):
    """实现基于耳切法的多边形三角化

    Parameters
    ----------
    pts
        多边形点集
    verbose
        是否输出辅助信息，包括顶点过多警告和顶点过多时显示进度条

    Returns
    -------
    ret
        array[len(pts) - 2, 3]，
        三角化后的面集合，每行指定一个三角形的三个顶点在pts中的下标

    Notes
    -----
        只支持简单多边形，即任意两条边的交点必须是顶点
    """
    n_vertices = pts.shape[0]
    polygon = ClippablePolygon(pts)
    vertices: VertexTriangle | None = None
    reflex_vertices: set[VertexTriangle] = set()
    triangles = np.empty((n_vertices - 2, 3), dtype=np.int64)

    for i in range(pts.shape[0]):
        vert = VertexTriangle(i, polygon)
        if vertices is None:
            vertices = vert
        else:
            vertices.append(vert)

        if vert.reflex:
            reflex_vertices.add(vert)

    rng = range(n_vertices - 2)

    if n_vertices > 200 and verbose:
        # TODO: 用c++重写？
        import warnings
        import tqdm

        warnings.warn(f'Applying ear-clipping on polygons'
                      f' with more than 200 vertices may lead to performance issue.'
                      f' (n_vertices = {n_vertices})'
                      f'\nProgress bar is automatically enabled,'
                      f' disable it by setting `verbose` = False.')
        rng = tqdm.tqdm(rng)

    for i in rng:
        best_to_remove: VertexTriangle | None = None
        for v in vertices:
            v: VertexTriangle

            if v.reflex:
                continue

            if v.test_is_ear_many(reflex_vertices):
                if best_to_remove is None:
                    best_to_remove = v
                    break

        if best_to_remove is None:
            # something wrong
            raise RuntimeError(f'Non-simple polygon is not supported yet.')

        triangles[i] = best_to_remove.clip()  # 从polygon中删除顶点

        prev_ = best_to_remove.prev
        next_ = best_to_remove.next

        best_to_remove.remove()
        if vertices == best_to_remove:
            vertices = next_

        prev_.assemble_triangle()
        next_.assemble_triangle()

        if best_to_remove.reflex:
            reflex_vertices.remove(best_to_remove)

        skip_reflex = len(reflex_vertices) == 0
        if prev_.reflex:
            reflex_vertices.add(prev_)
        elif not skip_reflex:
            try:
                reflex_vertices.remove(prev_)
            except KeyError:
                pass
            else:
                pass

        if next_.reflex:
            reflex_vertices.add(next_)
        elif not skip_reflex:
            try:
                reflex_vertices.remove(next_)
            except KeyError:
                pass
            else:
                pass

    return triangles



import torch
from SimPEG.simulation import BaseSimulation

from metalpy.mexin.injectors import extends
from metalpy.mexin import Mixin
from metalpy.mexin import Patch


class TorchContext(Mixin):
    def __init__(self, this, device='cuda'):
        super().__init__(this)
        self._device = device

    def post_apply(self, this):
        path = type(this).__module__
        dispatch = path.replace('SimPEG.', '').replace('.simulation', '')
        dispatch = f'metalpy.scab.extensions.{dispatch}.ext_torch'

        try:
            impl: Mixin = __import__(dispatch, fromlist=['TorchedImpl']).TorchedImpl
        except:
            print(f'Torch support for {path} is not implemented. Ignoring it.')
            return

        this.mixins.add(impl)

    def get_torch_device(self, this):
        return self._device

    @staticmethod
    def estimate_memory_cost(this, size=None):
        # size = self.survey.receiver_locations.shape[0] if size is None else size
        # return 100 * size * self.modelMap.shape[0]

        return None

    @staticmethod
    def estimate_batch_size(this, memory):
        return int(memory / this.estimate_memory_cost(1))


class Torched(Patch):
    def __init__(self, device='cuda'):
        super().__init__()
        self.device = device

    def apply(self):
        self.add_mixin(BaseSimulation, TorchContext, self.device)



import pandas as pd

from ..testing import ArrayOpTest


def make_indexed_dataframes(*arrs, columns=None):
    if columns is None:
        columns = range(0, len(arrs))
    for arr, col in zip(arrs, columns):
        yield pd.DataFrame(arr, columns=[col])


def make_dataframes(*arrs, indexed=False):
    if indexed:
        columns = None
    else:
        columns = [0] * len(arrs)
    yield from make_indexed_dataframes(*arrs, columns=columns)


def test_cstack():
    from metalpy.utils.array_ops import cstack
    ArrayOpTest(
        *make_indexed_dataframes([1, 2, 3], [4, 5, 6]),
        expects=pd.DataFrame([[1, 4], [2, 5], [3, 6]])
    ).test(cstack)


def test_rstack():
    from metalpy.utils.array_ops import rstack
    ArrayOpTest(
        *make_dataframes([1, 2, 3], [4, 5, 6]),
        expects=pd.DataFrame([1, 2, 3, 4, 5, 6])
    ).test(rstack)



import warnings

import numpy as np

from metalpy.utils.bounds import Bounds


def is_inside_cuboid(mesh, corner, lengths):
    warnings.warn('`is_inside_cuboid` is deprecated due to performance issue,'
                  ' use `is_inside_bounds` instead.')

    axes_mask = np.zeros(mesh.shape[1], dtype=bool)
    axes_mask[:len(corner)] = ~np.isnan(corner)
    axes_mask[:len(lengths)] &= ~np.isnan(lengths)
    corner_mask = axes_mask[:len(corner)]
    lengths_mask = axes_mask[:len(lengths)]

    if axes_mask.any():
        deltas = mesh[:, axes_mask] - corner[corner_mask]
        return np.all(deltas >= 0, axis=1) & np.all(deltas <= lengths[lengths_mask], axis=1)
    else:
        return np.full(mesh.shape[0], True)


def is_inside_bounds(mesh, bounds):
    bounds = Bounds(bounds)
    mask = np.ones(mesh.shape[0], dtype=bool)

    for a in range(bounds.n_axes):
        amin, amax = bounds.get(a)
        ax = mesh[:, a]
        if not np.isnan(amax):
            mask &= ax <= amax
        if not np.isnan(amin):
            mask &= ax >= amin

    return mask



from .array_ops import OperatorDispatcher, ArrayType

GetRow = OperatorDispatcher('GetRow')


def get_row(arr, ind):
    """从数组获取行

    Parameters
    ----------
    arr
        数组实例
    ind
        行索引或行索引数组 / 切片slice / 行索引或行索引数组（pandas）

    Returns
    -------
    ret
        返回获取的行，如果输入为切片范围或索引列表，则返回对应矩阵

    Notes
    -----
    对于Pandas，直接传入整数序列或切片时默认按行号索引，如果需要按索引取值，则需要通过 `pd.Index` 包装
    """
    return GetRow.dispatch(type(arr))(arr, ind)


@GetRow.register(ArrayType.numpy)
def _get_row_np(arr, ind):
    return arr[ind]


@GetRow.register(ArrayType.pandas)
def _get_row_pd(arr, ind):
    import pandas as pd
    if isinstance(ind, pd.Index) or getattr(ind, 'dtype', None) == bool:
        return arr.loc[ind]  # 行索引
    else:
        return arr.iloc[ind]  # 否则认为是行号



import abc

import numpy as np

from metalpy.scab.utils.misc import Field


class DemagnetizationSolver(abc.ABC):
    def __init__(
            self,
            receiver_locations: np.ndarray,
            xn: np.ndarray,
            yn: np.ndarray,
            zn: np.ndarray,
            base_cell_sizes: np.ndarray,
            source_field: Field
    ):
        self.receiver_locations = receiver_locations
        self.xn = xn
        self.yn = yn
        self.zn = zn
        self.base_cell_sizes = base_cell_sizes
        self.source_field = source_field

        self.model = None

    @property
    def kernel_type(self):
        return np.result_type(self.xn, self.receiver_locations)

    def dpred(self, model, source_field=None):
        """计算退磁效应下每个网格的三轴等效磁化率

        Parameters
        ----------
        model
            磁化率模型
        source_field
            外部场源，覆盖求解器定义时给定的场源信息。
            如果为None，则采用给定的默认场源信息

        Returns
        -------
        ret
            三轴等效磁化率矩阵
        """
        if source_field is None:
            source_field = self.source_field
            assert source_field is not None, ('`source_field` must be specified'
                                              ' either when initializing the solver(source_field=...)'
                                              ' or when calling dpred(source_field=...).')

        source_field = Field(source_field)
        if self.model is None or np.any(self.model != model):
            self.build_kernel(model)
            self.model = np.copy(model)

        nC = self.xn.shape[0]
        H0 = source_field.unit_vector
        H0 = np.tile(H0[None, :], nC).ravel()
        X = np.tile(model, 3).ravel()
        magnetization = X * H0

        return self.solve(magnetization)

    @abc.abstractmethod
    def build_kernel(self, model):
        """构造核矩阵
        """
        pass

    @abc.abstractmethod
    def solve(self, magnetization):
        """求解退磁效应下每个网格的三轴等效磁化率

        Parameters
        ----------
        magnetization
            磁化矩阵

        Returns
        -------
        ret
            三轴等效磁化率矩阵
        """
        pass



import numpy as np
import pandas as pd

from .. import is_array_like


def make_it_numpy(sth):
    if isinstance(sth, (pd.DataFrame, pd.Series)):
        return np.asarray(sth)
    else:
        return sth


class ArrayOpTest:
    def __init__(self, *args, expects=None, **kwargs):
        """构造针对ArrayOp的测试用例，
        只需要基于pandas目标提供参数，
        numpy目标下会自动将所有pandas目标进行转换后再执行测试

        Parameters
        ----------
        args, kwargs
            ArrayOp的参数
        expects
            期望的输出，如果给定为None，则会转换为pandas和numpy的等价性测试
            （测试两个目标下的输出是否一致，但不推荐）
        """
        self.args = args
        self.kwargs = kwargs
        self.expects = expects

    def test_pandas(self, op):
        got = op(*self.args, **self.kwargs)
        expected = self.expects

        if is_array_like(expected):
            return pd.testing.assert_frame_equal(
                got,
                expected
            )
        else:
            assert got == expected

    def test_numpy(self, op):
        args, kwargs = self.prepare_numpy()

        got = op(*args, **kwargs)
        expected = make_it_numpy(self.expects)

        if is_array_like(expected):
            return np.testing.assert_almost_equal(
                got,
                expected
            )
        else:
            assert got == expected

    def test_equivalence(self, op):
        args, kwargs = self.prepare_numpy()

        got = op(*args, **kwargs)
        expected = op(*self.args, **self.kwargs)
        expected = make_it_numpy(expected)

        if is_array_like(expected):
            return np.testing.assert_almost_equal(
                got,
                expected
            )
        else:
            assert got == expected

    def test(self, op):
        if self.expects is not None:
            self.test_pandas(op)
            self.test_numpy(op)
        else:
            self.test_equivalence(op)

    def prepare_numpy(self):
        args = [make_it_numpy(a) for a in self.args]
        kwargs = {k: make_it_numpy(v) for k, v in self.kwargs.items()}

        return args, kwargs



import numpy as np


def limit_significand(val, bits=None, tol=None):
    """将 val 转换为 IEEE754 格式浮点数可以准确表示的形式

    可选截断模式包含：

    - 按二进制有效位数截断
    - 按与原值的差距进行阶段

    Parameters
    ----------
    val
        需要限制有效位数的浮点数
    bits
        保留的二进制有效位数
    tol
        与原值的容差

    Returns
    -------
    truncated_val
        截断后的值
    """
    val = np.asarray(val)
    dt = val.dtype
    if np.issubdtype(dt, np.integer):
        return val

    finfo = np.finfo(dt)

    if bits is None and tol is None:
        tol = 10 ** -(finfo.precision // 2)

    if bits is None:
        bits = finfo.bits

    if tol is None:
        tol = -1

    int_part = val.astype(np.int64)
    float_part = val - int_part
    fp = 0
    a = 1
    for i in range(bits):
        sign = np.sign(float_part - fp)
        a /= 2
        if a <= tol:
            break
        else:
            fp += sign * a

    return int_part + fp




import os
import warnings
from pathlib import Path
from shutil import which


def check_gdal(cmd_name):
    if 'PROJ_LIB' not in os.environ:
        # 尝试修复找不到proj.db的问题
        # https://github.com/PDAL/PDAL/issues/2544

        # 如果是conda安装，有可能设置了GDAL_DATA但PROJ_LIB未设置
        if 'GDAL_DATA' in os.environ:
            # proj包和libgdal包同在`${env}/Library/share`目录下
            proj_lib_path = Path(os.environ['GDAL_DATA']) / '../proj'
            proj_lib_path = proj_lib_path.resolve()

            if proj_lib_path.exists():
                config_gdal(proj_lib=proj_lib_path)
            else:
                warnings.warn('`PROJ_LIB` not found in environment variables.'
                              ' Please check your `GDAL` installation.')
        else:
            warnings.warn('`GDAL_DATA` not found in environment variables.'
                          ' Please check your `GDAL` installation.')

    ret = which(cmd_name)
    if ret is None:
        raise RuntimeError('`GDAL` not found in PATH.'
                           ' `GDAL` can be installed with'
                           ' `conda install libgdal`'
                           ' or by following instructions from'
                           ' `https://gdal.org/download.html`.'
                           ' Or check to ensure `GDAL` binaries in PATH.')

    return ret


def config_gdal(proj_lib, gdal_data=None):
    os.environ['PROJ_LIB'] = os.fspath(proj_lib)
    if gdal_data is not None:
        os.environ['GDAL_DATA'] = os.fspath(gdal_data)



import numpy as np
import pytest

from metalpy.scab.modelling.shapes import Cuboid
from metalpy.scab.modelling.shapes.testing.shape3d_place_test import assert_placing_inside


def test_cuboid():
    assert_placing_inside(Cuboid(bounds=[1, 2, 3, 4, 5, 6]))


def test_cuboid_ctor():
    # 测试各种定义方式的等价性
    cubes = [
        Cuboid([0, 2, 0], [2, 0, 2]),
        Cuboid([0, 0, 0], [2, 2, 2]),
        Cuboid(origin=[0, 0, 0], size=[2, 2, 2]),
        Cuboid(origin=[0, 0, 0], size=2),
        Cuboid(origin=[0, 0, 0], center=[1, 1, 1]),
        Cuboid(end=[2, 2, 2], size=2),
        Cuboid(end=[2, 2, 2], center=[1, 1, 1]),
        Cuboid(center=[1, 1, 1], size=[2, 2, 2]),
        Cuboid(bounds=[0, 2, 0, 2, 0, 2]),
    ]

    origins = np.vstack([cube.origin for cube in cubes])
    lengths = np.vstack([cube.lengths for cube in cubes])

    assert np.all(origins == origins[0])
    assert np.all(lengths == lengths[0])


def test_cuboid_failure():
    # 测试严格模式下错误定义报错
    with pytest.raises(ValueError):
        _ = Cuboid([0, 2, 0], [2, 0, 2], no_corner_adjust=True)



from __future__ import annotations

import copy
import os.path
from typing import Any, Union, Iterable, cast, Mapping

import numpy as np
import tqdm
from discretize import TensorMesh

from metalpy.mepa import LinearExecutor
from metalpy.scab.utils.hash import dhash_discretize_mesh
from metalpy.utils.bounds import Bounds
from metalpy.utils.dhash import dhash
from metalpy.utils.file import ensure_dir, make_cache_directory
from metalpy.utils.model import as_pyvista_array
from .formats.osm import OSMFormat
from .formats.ptopo import PTopoFormat
from .layer import Layer
from .mix_modes import MixMode, Mixer
from .modelled_mesh import ModelledMesh
from .object import Object
from .shapes import Shape3D
from .shapes.full_space import FullSpace
from .shapes.shape3d import bounding_box_of
from ...utils.numeric import limit_significand


class Scene(OSMFormat, PTopoFormat):
    default_cache_dir = make_cache_directory('models')
    INACTIVE_BOOL = False
    INACTIVE_INT = 0
    INACTIVE_FLOAT = 0

    def __init__(self):
        self.objects_layer = Layer()
        self.backgrounds_layer = Layer(mix_mode=MixMode.KeepOriginal)

    @staticmethod
    def of(*shapes: Shape3D | Iterable[Shape3D],
           models: Union[Any, dict, list[Any], None] = None,
           skip_checking=False
           ) -> 'Scene':
        """从一组Shape3D实例创建场景，并赋予相同的模型值`models`

        Parameters
        ----------
        shapes
            一组Shape3D实例
        models
            三维几何体的参数
        skip_checking
            指示是否跳过对shapes的类型检查

        Returns
        -------
        scene
            新创建的场景

        Notes
        -----
        支持通过变长参数或数组传递shapes

        Examples
        --------
        >>> from metalpy.scab.modelling.shapes import Cuboid, Prism
        >>> shapes = [
        >>>     Cuboid([1, 1, 1], size=2),
        >>>     Prism([[0, 0], [-2, 0], [-2, 1], [-1, 1]], 1, 3),
        >>> ]
        >>> Scene.of(shapes)  # OK!
        >>> Scene.of(*shapes)  # OK!
        """
        ret = Scene()

        if not skip_checking:
            if len(shapes) == 1 and not isinstance(shapes[0], Shape3D):
                # 假设传入的为Shape3D实例数组，而非以变长参数形式传递
                shapes = shapes[0]

            assert all([isinstance(s, Shape3D) for s in shapes]), '`Scene` accepts only `Shape3D` instances.'

        if (
            models is None
            or isinstance(models, Mapping)
            or not isinstance(models, Iterable)
        ):
            models = [models] * len(shapes)

        for shape, model in zip(shapes, models):
            ret.append(shape, model)

        return ret

    def with_background(self, value, region_shape=None):
        self.append_background(value, region_shape)
        return self

    def append(self, 
               shape: Shape3D,
               models: Union[dict[str, Any], Any],
               mix_mode: Mixer = MixMode.Override) -> Object:
        """添加三维几何体

        Parameters
        ----------
        shape
            三维几何体
        models
            三维几何体的参数
        mix_mode
            对象混合模式

        Returns
        -------
        ret
            返回构造的三维几何体
        """
        obj = Object(shape, models, mix_mode=mix_mode)
        self.objects_layer.append(obj)

        return obj

    def extend(self,
               shapes: Iterable[Shape3D],
               models: Union[dict[str, Any], Any],
               mix_mode: Mixer = MixMode.Override) -> list[Object]:
        """添加多个三维几何体

        Parameters
        ----------
        shapes
            三维几何体集合
        models
            三维几何体的参数
        mix_mode
            对象混合模式

        Returns
        -------
        ret
            返回构造的三维几何体集合
        """
        return [self.append(s, models, mix_mode=mix_mode) for s in shapes]

    def append_background(self, value, region_shape=None):
        if region_shape is None:
            region_shape = FullSpace()
        obj = Object(region_shape, value)
        self.backgrounds_layer.append(obj)

        return obj

    @property
    def bounds(self):
        return bounding_box_of(self.shapes)

    def build_model(self, mesh,
                    executor=None,
                    progress=False,
                    cache=None,
                    cache_dir=None,
                    ) -> ModelledMesh:
        """在给定网格上构建模型

        Parameters
        ----------
        mesh
            用于构建模型的网格
        executor
            并行执行器
        progress
            是否显示进度条
        cache
            控制缓存文件，若为bool值，则指示是否启用缓存；若为str，则指示缓存文件的路径
        cache_dir
            控制缓存文件夹，指示放置缓存文件的文件夹，文件名会使用默认规则生成

        Returns
        -------
        ret
            若所有模型只包含默认键，则返回一个数组，为默认键下的值的构建结果

            否则返回字典，包含所有键以及该键下的值的构建结果

        Notes
        -----
            模型假设为非0值

            cache和cache_dir任意一个不为空时则会启动缓存，但如果cache和cache_dir同时被指定，则

            如果cache为True，则使用cache_dir与默认规则文件名

            如果cache为路径，则使用cache指定的文件路径
        """
        cache_filepath = self._determine_cache_filepath(mesh, cache, cache_dir)

        if cache_filepath is not None and os.path.exists(cache_filepath):
            import pickle
            with open(cache_filepath, 'rb') as f:
                # TODO: 添加额外的元信息判断来确保缓存有效？
                models_dict = pickle.load(f)
        else:
            if executor is None:
                executor = LinearExecutor(1)

            layers = tuple(self.layers)
            n_workers = executor.get_n_workers()

            if progress:
                n_tasks = sum((obj.n_tasks for layer in layers for obj in layer))
                progress = executor.progress(total=n_tasks * n_workers)
            else:
                progress = None

            mesh_centers = np.asarray(mesh.cell_centers)

            shuffle = n_workers > 1
            mesh_alloc = executor.arrange(mesh_centers, shuffle=shuffle)

            futures = executor.distribute(self._build_mesh_worker, layers, mesh_alloc, progress=progress)
            results = executor.gather(futures)

            models_dict = {}
            for key in results[0].keys():
                # 组装并还原被打乱的数据
                models_dict[key] = mesh_alloc.reassemble([result[key] for result in results])

            if cache_filepath is not None:
                import pickle
                with open(cache_filepath, 'wb') as f:
                    pickle.dump(models_dict, f)

            if progress is not None:
                progress.close()

        return ModelledMesh(mesh, models_dict)

    def create_mesh(self, cell_size=None, n_cells=None, bounds=None) -> TensorMesh:
        """根据场景边界构建网格

        Parameters
        ----------
        cell_size : number or array(3,)
            网格长宽高相等或分别定义网格x, y, z方向宽度大小
        n_cells : number or array(3,)
            总网格数或x, y, z方向网格数
        bounds : array(6,) or Bounds
            网格范围，为array(6,)[xmin, xmax, ymin, ymax, zmin, zmax]或Bounds实例，为None的位置会使用默认值

        Returns
        -------
        ret
            构建的网格

        Notes
        -----
            原点是场景的边界的最小值。
            若指定cell_size，则边界点保证大于场景边界；
            若指定n_cells，则边界点是场景的边界的最大值。

            若n_cells为单个值，则用于指定总网格数，保证生成的网格数小于等于该值。
        """
        actual_bounds: Bounds = self.bounds
        if bounds is not None:
            actual_bounds.override(by=bounds, inplace=True)

        bounds = actual_bounds
        sizes = bounds.extent

        if cell_size is not None:
            if np.ndim(cell_size) == 0:  # array 状态下的int不会被视为 scalar，不能使用np.isscalar
                cell_size = [cell_size] * 3
            cell_size = np.asarray(cell_size)
            n_cells = np.ceil(sizes / cell_size).astype(int)
        else:
            if not isinstance(n_cells, Iterable):
                avg_grids = (n_cells / np.prod(sizes)) ** (1 / 3)
                n_cells = (avg_grids * sizes).astype(int)
                cell_size = [1 / avg_grids] * 3
            else:
                n_cells = np.asarray(n_cells)
                cell_size = sizes / n_cells

        return TensorMesh([[(d, n)] for d, n in zip(cell_size, n_cells)], origin=bounds[::2])

    def build(self, cell_size=None, n_cells=None, bounds=None,
              executor=None, progress=False,
              cache=None, cache_dir=None,
              ) -> ModelledMesh:
        """根据给定的网格尺寸，构建场景的网格和模型，是create_mesh和build_model的组合

        Parameters
        ----------
        cell_size : number or array(3,)
            网格长宽高相等或分别定义网格x, y, z方向宽度大小
        n_cells : number or array(3,)
            总网格数或x, y, z方向网格数
        bounds : array(6,) or Bounds
            网格范围，为array(6,)[xmin, xmax, ymin, ymax, zmin, zmax]或Bounds实例，为None的位置会使用默认值
        executor
            并行执行器
        progress
            是否显示进度条
        cache
            指示缓存行为，详见Scene.build_model。单独使用cache=True来使用默认缓存路径或cache='...'来指定缓存文件路径
        cache_dir
            指示缓存行为，详见Scene.build_model。单独使用cache_dir='...'来指定缓存所在的文件夹路径

        Returns
        -------
        model_mesh
            模型网格

        See Also
        --------
        Scene.create_mesh : 构造网格
        Scene.build_model : 构造模型
        """
        mesh = self.create_mesh(cell_size=cell_size, n_cells=n_cells, bounds=bounds)
        pruned = self.build_model(mesh, executor=executor, progress=progress, cache=cache, cache_dir=cache_dir)

        return pruned

    def build_active(
            self,
            active_cells,
            cell_size=None,
            iters=1,
            limit_sig=True
    ) -> ModelledMesh:
        """重新构建网格和模型，使得有效网格数接近 `active_cells`

        Parameters
        ----------
        active_cells
            有效网格数
        cell_size
            网格尺寸，若指定，则会以该尺寸作为网格比例，否则使用各向同性网格
        iters
            迭代次数，每次迭代会根据上一次的结果调整网格尺寸
        limit_sig
            是否限制网格尺寸的有效数字位数

        Returns
        -------
        model_mesh
            模型网格

        See Also
        --------
        Scene.build : 构造网格和模型
        """
        if cell_size is None:
            model_mesh = self.build(n_cells=active_cells)
            cell_size = [h[0] for h in model_mesh.mesh.h]
        else:
            model_mesh = self.build(cell_size=cell_size)

        for i in range(1, iters + 1):
            ratio = (model_mesh.n_active_cells / active_cells) ** (1 / 3)
            cell_size = [c * ratio for c in cell_size]
            if i == iters and limit_sig:
                cell_size = limit_significand(cell_size)
            model_mesh = self.build(cell_size=cell_size)

        return model_mesh

    def build_objects(self, mesh, progress=None):
        if progress is True:
            progress = tqdm.tqdm(total=len(self))

        models = []

        for obj, ind, sub_models in Scene._build_objects(self.objects, mesh, progress):
            mask = Scene.is_active(ind)
            models.append(ModelledMesh(mesh, dict(sub_models), ind_active=mask))

        return models

    @property
    def layers(self) -> Iterable[Layer]:
        yield self.objects_layer
        yield self.backgrounds_layer

    @property
    def objects(self):
        for layer in self.layers:
            for obj in layer:
                yield obj

    @property
    def shapes(self) -> Iterable[Shape3D]:
        for obj in self.objects:
            yield obj.shape

    def filter(self, model):
        for obj in self:
            if model in obj.models:
                yield obj

    def __len__(self):
        return sum((len(layer) for layer in self.layers))

    def __iter__(self) -> Iterable[Object]:
        for obj in self.objects:
            yield obj

    def __getitem__(self, item) -> Object:
        if item < 0:
            item = len(self) + item

        for layer in self.layers:
            size = len(layer)
            if item < size:
                return layer[item]
            else:
                item -= size

    def to_multiblock(self):
        import pyvista as pv

        ret = pv.MultiBlock()
        for obj in self.objects:
            poly = obj.to_polydata()
            if poly is not None:
                ret.append(poly)

        return ret

    def plot(self, *args, **kwargs):
        """直接生成PyVista对象并调用plot在窗口中绘制

        Parameters
        ----------
        kwargs
            其它适用于pv.MultiBlock.plot的参数

        See Also
        --------
            Scene.to_multiblock : 生成PyVista对象
        """
        self.to_multiblock().plot(*args, **kwargs)

    @staticmethod
    def mesh_to_polydata(mesh, models: Union[np.ndarray, dict[str, np.ndarray]]):
        """将给定网格转换为对应的PyVista网格实例，并绑定模型值

        Parameters
        ----------
        mesh
            待绑定的网格
        models
            待绑定的模型

        Returns
        -------
        ret
            绑定给定网格和模型的pyvista网格对象

        Notes
        -----
        虽然名字为mesh_to_polydata，但他的返回值并不是pv.PolyData而是pv.RectilinearGrid
        """
        import pyvista as pv

        if not isinstance(models, dict):
            models = {'active': models}
        else:
            models = copy.copy(models)

        for key in models:
            models[key] = as_pyvista_array(models[key])

        # TensorMesh会转换为pv.RectilinearGrid
        grids: pv.RectilinearGrid = cast(
            pv.RectilinearGrid,
            mesh.to_vtk(models=models)
        )

        for key in models:
            # 将第一个model设为active
            grids.set_active_scalars(key)
            break

        return grids

    def __dhash__(self):
        return dhash(*self.layers)

    @staticmethod
    def get_inactive_value_for_type(dtype):
        if dtype == bool:
            return Scene.INACTIVE_BOOL
        elif dtype == np.issubdtype(dtype, np.integer):
            return Scene.INACTIVE_INT
        else:
            return Scene.INACTIVE_FLOAT

    @staticmethod
    def is_active(model):
        # TODO: 考察是否有必要视0为非活动网格
        return model if model.dtype == bool else model != 0

    @staticmethod
    def compute_active(models: Iterable[np.ndarray]):
        ind_active = None
        for model in models:
            if ind_active is None:
                ind_active = Scene.is_active(model)
            else:
                ind_active |= Scene.is_active(model)

        return ind_active

    @staticmethod
    def _merge_models(models, new_models: Iterable[tuple[str, np.ndarray]], mask, mixer):
        """将new_models合并到models中，inplace

        Parameters
        ----------
        models
            已有的model
        new_models
            将要合并的model，以迭代器形式提供
        mask
            新model的有效格mask，如果指定为None，则会为每个待合并model单独计算mask
        mixer
            混合器

        Notes
        -----
            会直接修改models以及其内的数组
        """
        for key, current_layer in new_models:
            if mask is not None:
                current_mask = mask
            else:
                current_mask = Scene.is_active(current_layer)

            if key not in models:
                models[key] = current_layer
            else:
                prev_layer = models[key]
                filled_ind = Scene.is_active(prev_layer)

                # 将已有的层转换为可以容纳待合并各层数据的类型（如int -> float）
                promoted_type = np.promote_types(prev_layer.dtype, current_layer.dtype)
                if promoted_type != prev_layer.dtype:
                    prev_layer = models[key] = prev_layer.astype(promoted_type)

                overlapping_mask = filled_ind & current_mask
                non_overlapping_mask = current_mask ^ overlapping_mask

                prev_layer[overlapping_mask] = mixer(
                    prev_layer[overlapping_mask],
                    current_layer[overlapping_mask]
                )
                prev_layer[non_overlapping_mask] = current_layer[non_overlapping_mask]

    @staticmethod
    def _build_objects(objects: objects, mesh, progress=None):
        models = {}

        for obj in objects:
            shape = obj.shape

            # place的结果应该为布尔数组或范围为[0, 1]的数组，指示对应网格位置是否有效或有效程度
            ind: np.ndarray = shape.place(mesh, progress)

            def model_generator():
                for key, current_value in obj.items():
                    if ind.dtype == bool:
                        current_layer = np.zeros_like(ind, dtype=type(current_value))
                        current_layer[ind] = current_value
                    else:
                        current_layer = ind * current_value
                    yield key, current_layer

            yield obj, ind, model_generator()

            if progress is not None and not obj.progress_manually:
                progress.update(obj.n_tasks)

        return models

    @staticmethod
    def _build_layer(layer: Layer, mesh, progress=None):
        objects = layer.objects
        models = {}

        for obj, ind, sub_models in Scene._build_objects(objects, mesh, progress):
            mask = Scene.is_active(ind)
            Scene._merge_models(models, sub_models, mask, obj.mixer)

        return models

    @staticmethod
    def _build_mesh_worker(layers: list[Layer], mesh, progress):
        models = {}

        for layer in layers:
            layer_models = Scene._build_layer(layer, mesh, progress)
            Scene._merge_models(models, layer_models.items(), None, layer.mixer)

        return models

    def _determine_cache_filepath(self, mesh, cache, cache_dir):
        cache_filepath = None
        if isinstance(cache, bool) and cache:
            cache_filename = self._generate_model_filename(mesh)
            if cache_dir is not None:
                cache_filepath = os.path.join(cache_dir, cache_filename)
            else:
                ensure_dir(Scene.default_cache_dir)
                cache_filepath = os.path.join(os.path.abspath(Scene.default_cache_dir), cache_filename)
        elif isinstance(cache, str):
            cache_filepath = cache
        elif cache is None:
            if cache_dir is not None:
                cache_filename = self._generate_model_filename(mesh)
                cache_filepath = os.path.join(cache_dir, cache_filename)

        return cache_filepath

    def _generate_model_filename(self, mesh: TensorMesh):
        mesh_hash = dhash_discretize_mesh(mesh).hexdigest(digits=6)
        model_hash = dhash(self).hexdigest(digits=6)

        origin = mesh.origin.astype(float)
        lengths = np.asarray([h.sum() for h in mesh.h])

        components = []

        for axis, x0, length in zip('xyz', origin, lengths):
            left = f"{x0:f}".rstrip("0").rstrip(".")
            right = f"{x0 + length:f}".rstrip("0").rstrip(".")
            components.append(f'{axis}({left}~{right})')

        avg_node_size = lengths / [h.shape[0] for h in mesh.h]
        if np.all(avg_node_size - avg_node_size[0] < 1e-7):
            avg_node_size = np.asarray(avg_node_size[0])
            components.append(f'avg_gs({avg_node_size:.2})')
        else:
            components.append(f'avg_gs{(*np.around(avg_node_size, decimals=2),)}')

        components.append(f'mesh({mesh_hash})')
        components.append(f'model({model_hash})')

        return '#'.join(components) + '.pkl'



import numpy as np

from metalpy.utils.dhash import dhash
from metalpy.utils.bounds import Bounds
from . import Shape3D


class Ellipsoid(Shape3D):
    def __init__(self, a, b, c):
        """定义一个椭球体，球心在原点

        Parameters
        ----------
        a
            椭球体在x方向上的半轴长
        b
            椭球体在y方向上的半轴长
        c
            椭球体在z方向上的半轴长
        """
        super().__init__()
        self.radii = np.asarray([a, b, c])

    @staticmethod
    def sphere(r) -> 'Ellipsoid':
        """定义一个球体

        Parameters
        ----------
        r
            半径

        Returns
        -------
        ret
            半径为r的球体
        """
        return Ellipsoid(r, r, r)

    @staticmethod
    def spheroid(a, c, polar_axis=2) -> 'Ellipsoid':
        """定义一个旋转椭球体（三个轴长为a, a, c）

        a > c时，其为扁椭球体

        a < c时，其为长椭球体

        默认polar_axis = 2时，其极轴在z轴上，方程为
            (x^2 + y^2) / a^2 + z^2 / c^2 = 1

        Parameters
        ----------
        a
            equatorial，半赤道轴长
        c
            polar，半极轴长
        polar_axis
            极轴方向，0 - x，1 - y，2 - z

        Returns
        -------
        ret
            中心点在原点，半赤道轴长为a，半极轴长为c，极轴为polar_axis的旋转椭球体
        """
        r = [a, a, a]
        r[polar_axis] = c
        return Ellipsoid(*r)

    @staticmethod
    def ellipsoid(a, b, c) -> 'Ellipsoid':
        """同直接调用构造函数
        """
        return Ellipsoid(a, b, c)

    def do_place(self, mesh_cell_centers, progress):
        return np.sum((mesh_cell_centers / self.radii) ** 2, axis=1) < 1

    def __dhash__(self):
        return dhash(super().__dhash__(), *self.radii)

    def do_clone(self, deep=True):
        return Ellipsoid(*self.radii)

    @property
    def local_bounds(self):
        return Bounds(np.c_[-self.radii, self.radii].ravel())

    def to_local_polydata(self):
        import pyvista as pv

        ret = pv.ParametricEllipsoid(*self.radii)
        ret.flip_normals()

        return ret

    @property
    def volume(self):
        return np.pi * 4 / 3 * np.prod(self.radii)

    @property
    def area(self):
        return np.pi * 4 / 3 * (self.radii * np.roll(self.radii, 1)).sum()



from .simpeg_patch_context import simpeg_patched

from .tied import Tied
from .distributed import Distributed
from .progressed import Progressed
from .demag import Demaged

# try:
#     from .torched import Torched
# except ImportError:
#     pass  # most likely torch is not installed



"""
Notes
-----
为了防止多余的import占用运行时间，因此保持只在必要时import相关类型，
但是缺陷是每次添加新的Builder时，需要进行如下三步操作：

- `TYPE_CHECKING` 中引入类型
- `SimulationBuilder` 中定义 `of` 的重载
- 使用__register_builder注册实际类型

其中前两步用于注解`SimulationBuilder.of`的返回值，帮助ide提示和类型检查，
第三步为实际注册对应Simulation的builder类。

实现原理
-------
- `supplier` 用于直接绑定实参，多次调用会覆盖。

实现方式为使用 `_supplies` 注解，
`_supplies` 注解的函数，会自动将返回值绑定到类构造函数的实参，

- `assembler` 用于构造指定实参，在实际构造对象时，对于所有未绑定实参的必选参数，会搜索调用其对应的assembler构造实参。

实现方式为使用 `_assembles` 注解，
`_assembles` 会将函数注册为类构造函数指定形参的assembler以供搜索调用。
"""
from __future__ import annotations

import copy
import functools
import inspect
import types
import warnings
from functools import lru_cache
from typing import overload, TYPE_CHECKING, cast, Iterable

from metalpy.scab import simpeg_patched

from metalpy.mexin import Patch

from metalpy.mexin.utils import TypeMap
from metalpy.mexin.utils.misc import reget_object
from metalpy.utils.arg_specs import ArgSpecs
from metalpy.utils.object_path import ObjectPath, get_full_qualified_path
from metalpy.utils.string import format_string_list
from metalpy.utils.type import undefined
from .utils import MissingArgsException, MissingArgException

if TYPE_CHECKING:
    # 用于类型检查时前向声明，运行时`TYPE_CHECKING` == False，import代码不会执行
    from SimPEG.potential_fields import magnetics
    from .potential_fields import magnetics as _magnetics


class SimulationBuilder:
    _registry = TypeMap(allow_match_parent=True)
    _all_assemblers = {}
    _all_suppliers = {}

    def __init__(self, sim_cls):
        """用于将任意对象的构建包装为builder模式，
        并通过提供合理的默认值来简化对象创建流程

        Parameters
        ----------
        sim_cls
            需要包装的类

        Notes
        -----
        一般只用于包装已有的如第三方库中的类，
        如果自己实现新的类的话，有必要的话直接实现成builder模式就好了，没必要再包装一层。
        """
        self.sim_cls = sim_cls
        self.args = ArgSpecs.from_class_mro(self.sim_cls)
        self._assemblers = {}
        self._patches: list[Patch] = []

    def patched(self, *patches: Patch):
        self._patches.extend(patches)
        return self

    def build(self):
        with simpeg_patched(*self._patches):
            for name, assembler in self.get_assemblers():
                key = self.args.find_arg_key(name)
                if key is None:
                    continue

                try:
                    self.args.get_bound_arg(key)
                except KeyError:
                    self.args.bind_arg(key, assembler())

            try:
                ret = self.args.call(reget_object(self.sim_cls))
                return ret
            except ValueError as e:
                ex = MissingArgsException()
                for arg in e.args[1] + e.args[2]:
                    ex.append(self.report_missing(arg.name))
                raise ex

    def report_missing(self, name, methods=None):
        return MissingArgException(name, methods if methods is not None else self.get_suppliers(name))

    def __new__(cls, sim_cls: type):
        builder_cls = cast(type[SimulationBuilder], SimulationBuilder._registry.get(sim_cls))
        if builder_cls is None:
            raise NotImplementedError(f'SimulationBuilder does not support {sim_cls.__name__} for now.')
        ret = super(SimulationBuilder, cls).__new__(builder_cls)
        return ret

    @staticmethod
    @overload
    def of(sim_cls: 'type[magnetics.Simulation3DIntegral]') -> '_magnetics.Simulation3DIntegralBuilder':
        ...

    @staticmethod
    def of(sim_cls):
        return SimulationBuilder(sim_cls)

    @staticmethod
    def _implies(*keysets):
        keysets = tuple((keyset,) if isinstance(keyset, str) else keyset for keyset in keysets)

        def decorator(func):
            cls_name = str(ObjectPath.of(func).parent)
            suppliers = SimulationBuilder._all_suppliers.setdefault(cls_name, {})

            name = getattr(func, '__name__', None)
            if name is None:
                warnings.warn(f'Unnamed function will not be notified if missing: {ObjectPath.of(func)}.')
            else:
                for keyset in keysets:
                    for key in keyset:
                        suppliers.setdefault(key, []).append(func.__name__)

            return func

        return decorator

    @staticmethod
    def _supplies(keyset0: str | Iterable[str],
                  *keysets: str | Iterable[str],
                  allow_vars=False,
                  pass_current_args=False):
        """指示该函数为对应参数的提供器。用户调用该函数时会为对应形参提供实参。
        所修饰的函数会通过返回值为给定参数名提供参数，并会被包装为返回self的链式调用。

        所修饰的函数需要满足以下条件：

        - 返回值绑定到keysets指定的参数上，当keysets长度为1时，返回单个值，否则返回一个长度与keysets一致的元组；
        - 在对应位置返回metalpy.utils.type.undefined可以指定该参数不绑定值；
        - 如果指定`pass_current_args`为True，则函数除了接受自身需要的参数，还需要允许接受以keysets指定参数名传入的先前绑定的参数（如果存在），
          这可以通过指定对应名字的带默认参数的关键字参数（如model_type=None）或者变关键字参数（**kwargs）。

        Parameters
        ----------
        keyset0, keysets
            给定关键字，长度与返回值数匹配，每个元素可以是单个或者若干个参数名。
            指示将对应位置的返回值绑定到对应的所有参数名下。
        allow_vars
            指示是否允许在目标函数固定参数中找不到对应参数时，将其绑定到目标的varkw上，默认为False，忽略该参数
        pass_current_args
            指示是否在调用函数时，同时传入当前参数已绑定的值（如果存在）

        Returns
        -------
        wrapper
            返回的组装器函数包装器，负责将修饰的函数注册为对应参数的提供者

        Notes
        -----
        如果找不到给定的参数名，且allow_vars为False或目标不接受varkw，则抛出错误

        Examples
        --------
        例如设置一个函数为名为`model_type`的形参提供实参，
        则如果之前该实参有绑定过值，且`pass_current_args`为True，
        _supplies的wrapper会将该值连同用户输入一并传递给绑定的函数，
        因此被包装的函数需要提供对应名字的参数（model_type=None）或接受关键字变参（**_ 或 **kwargs）。

        >>>@SimulationBuilder._supplies('model_type', pass_current_args=True)
        >>>def model_type(self, mtype=None, *, model_type=None):
        >>>    return mtype

        >>>builder.model_type('scalar')  # model_type的输入: type='scalar'
        >>>builder.model_type('vector')  # model_type的输入: type='vector', model_type='scalar'
        """
        keysets = (keyset0, *keysets)
        keysets = tuple((keyset,) if isinstance(keyset, str) else keyset for keyset in keysets)

        def decorator(func):
            SimulationBuilder._implies(*keysets)(func)

            @functools.wraps(func)
            def wrapper(self, *args, **kwargs):
                if pass_current_args:
                    for keyset in keysets:
                        for key in keyset:
                            try:
                                val = self.args.get_bound_arg(key)
                                kwargs[key] = val
                            except KeyError:
                                pass

                ret = func(self, *args, **kwargs)
                if len(keysets) == 1:
                    # 如果指定绑定的参数长度为1，则返回值不要求为形似 `return x,` 的元组形式
                    ret = (ret,)

                assert len(ret) == len(keysets), \
                    "Arg suppliers' return values must have same length as `_supplies` signature."

                for keyset, val in zip(keysets, ret):
                    if undefined == val:
                        continue
                    bound_at_least_one = False
                    for key in keyset:
                        try:
                            key = self.args.find_arg_key(key, allow_vars=allow_vars)
                            self.args.bind_arg(key, val)
                            bound_at_least_one = True
                        except KeyError:
                            pass

                    if not bound_at_least_one:
                        warnings.warn(f'Failed to bind argument for'
                                      f' {format_string_list(keyset, multiline=True, quote="`")}')

                return self

            return wrapper

        return decorator

    @staticmethod
    def supplier(
        keyset: str | Iterable[str],
        allow_vars=False
    ):
        """快速创建一个直接传递（passthrough）的supplier

        Parameters
        ----------
        keyset
            给定关键字，长度与返回值数匹配，可以是单个或者若干个参数名。
        allow_vars
            指示是否允许在目标函数固定参数中找不到对应参数时，将其绑定到目标的varkw上，默认为False，忽略该参数
        """
        @SimulationBuilder._supplies(keyset, allow_vars=allow_vars)
        def supplier(_, arg):
            return arg

        return supplier

    def bind_supplier(
        self,
        keyset: str | Iterable[str],
        allow_vars=False
    ):
        """运行时创建supplier，并绑定到当前实例

        Parameters
        ----------
        keyset
            给定关键字，长度与返回值数匹配，可以是单个或者若干个参数名。
        allow_vars
            指示是否允许在目标函数固定参数中找不到对应参数时，将其绑定到目标的varkw上，默认为False，忽略该参数
        """
        if isinstance(keyset, str):
            keyset = (keyset,)

        for key in keyset:
            setattr(self, key, types.MethodType(
                SimulationBuilder.supplier(keyset, allow_vars=allow_vars),
                self
            ))

    @staticmethod
    def _assembles(key0: str, *other_keys: str):
        """指示该函数为对应参数的组装器，在指定形参缺失时由Builder系统调用该方法构造实参。

        所修饰的函数需要满足以下条件：

        - 返回值有且仅能有一个；
        - 只在keys指定的任意一个形参`P`缺失实参时调用，返回值直接绑定到缺失`P`上。

        Parameters
        ----------
        keys, other_keys
            给定关键则，长度与返回值数匹配，每个元素可以是单个或者若干个参数名。
            指示将对应位置的返回值绑定到对应的所有参数名下。

        Returns
        -------
        wrapper
            返回的组装器函数包装器，负责将修饰的函数注册为对应参数的组装器
        """
        keys = (key0, *other_keys)

        def decorator(func):
            cls_name = str(ObjectPath.of(func).parent)
            assemblers = SimulationBuilder._all_assemblers.setdefault(cls_name, {})

            for key in keys:
                assemblers[key] = func.__name__

            return func

        return decorator

    def get_assemblers(self):
        for cls in reversed(inspect.getmro(type(self))):
            cls_name = get_full_qualified_path(cls)
            for key, assembler_name in SimulationBuilder._all_assemblers.get(cls_name, {}).items():
                yield key, getattr(self, assembler_name)

    @classmethod
    @lru_cache
    def get_suppliers(cls, name):
        suppliers = []
        for sub_cls in reversed(inspect.getmro(cls)):
            cls_name = get_full_qualified_path(sub_cls)
            suppliers.extend(SimulationBuilder._all_suppliers.get(cls_name, {}).get(name, tuple()))

        return suppliers

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls, self.sim_cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        # ref:
        # https://stackoverflow.com/questions/1500718/how-to-override-the-copy-deepcopy-operations-for-a-python-object
        cls = self.__class__
        result = cls.__new__(cls, self.sim_cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, copy.deepcopy(v, memo))
        return result


def __register_builder(key_cls_path):
    def decorator(func):
        SimulationBuilder._registry.map(key_cls_path, func)
        return func
    return decorator


@__register_builder('SimPEG.potential_fields.magnetics.simulation.Simulation3DIntegral')
def _():
    from .potential_fields.magnetics.simulation import Simulation3DIntegralBuilder
    return Simulation3DIntegralBuilder



import warnings

import numpy as np

from metalpy.utils.dhash import dhash
from . import InfiniteShape


class FullSpace(InfiniteShape):
    def __init__(self):
        """定义一个覆盖整个空间的几何体
        """
        super().__init__()

    def do_place(self, mesh_cell_centers, progress):
        indices = np.ones(mesh_cell_centers.shape[0], dtype=bool)
        return indices

    def __dhash__(self):
        return dhash(0)

    def do_clone(self, deep=True):
        return FullSpace()

    def to_local_polydata(self):
        return None

    @property
    def volume(self):
        warnings.warn(f'Trying to access `volume` of `{FullSpace.__name__}`.')
        return 0

    @property
    def area(self):
        warnings.warn(f'Trying to access `area` of `{FullSpace.__name__}`.')
        return 0



from metalpy.mepa import Worker, Executor


class Distributable:
    def __init__(self):
        pass

    def should_distribute_to(self, worker: Worker):
        return True

    def distribute_to(self, executor: Executor, worker: Worker):
        if self.should_distribute_to(worker):
            return self
        else:
            return None



from metalpy.mepa import Worker
from .distributable import Distributable


class DistributeOnce(Distributable):
    def __init__(self):
        super().__init__()

    def should_distribute_to(self, worker: Worker):
        return worker.get_in_group_id() == 0



from __future__ import annotations

from typing import Union

import tqdm
from SimPEG.simulation import BaseSimulation

from metalpy.mepa import Executor, Worker, ParallelProgress
from metalpy.mexin import Mixin
from metalpy.mexin.patch import Patch
from .distributed.policies import Distributable

ProgressContext = Union[tqdm.tqdm, ParallelProgress, None]


class Progress(Mixin):
    def __init__(self, this: BaseSimulation, patch: Progressed):
        super().__init__(this)
        self.patch = patch

    def get_progress_ctx(self, _=None):
        return self.patch.get_context()

    def reset(self, _, total):
        self.patch.reset(total)

    def update(self, _, n):
        self.get_progress_ctx().update(n)

    def close(self, _):
        self.get_progress_ctx().close()


class Progressed(Patch, Distributable):
    BaseProgressStyles = {
        'unit_scale': True,
        'smoothing': 0.1,
    }

    def __init__(self, _progress_ctx: ProgressContext = None, **tqdm_kw):
        """进度条插件，为正演添加进度条机制（目前）

        Parameters
        ----------
        _progress_ctx
            进度上下文，用户不应手动指定
        tqdm_kw
            用于 `tqdm` 的进度条参数
        """
        super().__init__()
        self.progress_ctx: ProgressContext = _progress_ctx
        self.tqdm_kw = {**Progressed.BaseProgressStyles, **tqdm_kw}

    def apply(self):
        if len(self.context.get_patches()) == 1:
            raise RuntimeError("Progressed patch cannot be used alone due to SimPEG's design")
        self.add_mixin(BaseSimulation, Progress, patch=self)

    def get_context(self, executor: Executor | None = None, allow_restart=False):
        if self.progress_ctx is None or (self.progress_ctx.disable and allow_restart):
            if executor is not None:
                self.progress_ctx = executor.progress(**self.tqdm_kw)
            else:
                self.progress_ctx = tqdm.tqdm(**self.tqdm_kw)
        return self.progress_ctx

    def reset(self, total):
        # 允许重启进度条以支持重用 Progressed 实例
        self.get_context(allow_restart=True).reset(total)

    def distribute_to(self, executor: Executor, worker: Worker):
        # 允许重启进度条以支持重用 Progressed 实例
        return Progressed(_progress_ctx=self.get_context(executor, allow_restart=True))



from typing import Union, Iterable

import numpy as np
import taichi as ti
import tqdm

from metalpy.utils.taichi import (
    ti_size_max,
    ti_kernel,
    ti_cfg,
    ti_ndarray,
    ti_func,
    ti_data_oriented,
    ti_ndarray_from,
    ensure_contiguous
)
from metalpy.utils.taichi_kernels import ti_index
from metalpy.utils.type import ensure_as_iterable, not_none_or_default
from ...value_observer import ValueObserver


class Receiver:
    def __init__(self,
                 receiver_locations: np.ndarray,
                 components: Union[str, Iterable[str]]):
        self.receiver_locations = receiver_locations
        self.components = components

    @property
    def n_rows(self):
        return self.receiver_locations.shape[0] * len(self.components)


@ti_data_oriented
class TaichiSimulation3DIntegral:
    MType_Vector = 0
    MType_Scalar = 1
    NonAccumulative = 0
    Accumulative = 1
    Layout_SoA = ti.Layout.SOA
    Layout_AoS = ti.Layout.AOS

    def __init__(self,
                 receivers: Union[Receiver, list[Receiver]],
                 xn: np.ndarray,
                 yn: np.ndarray,
                 zn: np.ndarray,
                 base_cell_sizes: np.ndarray,
                 model_type: int = MType_Scalar,
                 row_stype: int = Layout_SoA,
                 col_stype: int = Layout_AoS,
                 tmi_projection: np.ndarray = None,
                 magnetization: np.ndarray = None,):
        """用于计算磁场的积分正演

        Parameters
        ----------
        receivers
            接收点信息包含坐标和待求解分量，
            坐标应为array(n_obs, 3)，
            待求解分量可选 'tmi', 'bx', 'by', 'bz', 'bxx', 'bxy', 'bxz', 'byy', 'byz', 'bzz'
        xn, yn, zn
            网格边界坐标，均为array(n_cells, 2)
        base_cell_sizes
            基础单元大小，array(3,)
        model_type
            模型类型，可选MType_Scalar或MType_Vector
        row_stype
            核矩阵行存储类型/正演算子的输入类型，可选Layout_SoA或Layout_AoS
        col_stype
            核矩阵列存储类型/正演算子的输出类型，可选Layout_SoA或Layout_AoS
        tmi_projection
            tmi投影矩阵，array(3,)，在不需要计算tmi分量时可不传
        magnetization
            磁化矩阵，array(n_cells, 3)，默认为全1
        """
        receivers: list[Receiver] = ensure_as_iterable(receivers)

        if tmi_projection is None:
            for receiver in receivers:
                assert 'tmi' not in receiver.components, "'tmi_projection' is required when computing tmi"
            tmi_projection = np.asarray([0])  # dummy variable

        if magnetization is None:
            magnetization = 1

        magnetization = np.asarray(magnetization)
        if magnetization.ndim == 2 and np.allclose(magnetization - magnetization[0], 0):
            # 标量模式下SimPEG会将地磁场重复n_cell行，判断如果重复则可以压缩磁化矩阵
            magnetization = magnetization[0]

        self.receivers = receivers
        self.xn = xn
        self.yn = yn
        self.zn = zn
        self.n_cells = xn.shape[0]
        self.base_cell_sizes = base_cell_sizes
        self.magnetization = magnetization
        self.tmi_projection = tmi_projection
        self.row_stype = row_stype
        self.col_stype = col_stype
        self.model_type = model_type

    def dpred(self,
              model: np.ndarray = None,
              progress: tqdm.tqdm = None):
        """计算正演结果

        Parameters
        ----------
        model
            模型参数
        progress
            tqdm进度条对象

        Returns
        -------
        ret
            正演结果

        Notes
        -----
            如果model_type==MType_Scalar，则model应为array(n_cells,)

            如果model_type==MType_Vector，则model应为array(n_cells, 3)

            如果model为None，则返回核矩阵。如果为scalar模型，核矩阵列数为n_cells，如果为vector模型，核矩阵列数为3*n_cells

            如果model不为None，则返回正演结果列向量
        """
        is_cpu = ti_cfg().arch == ti.cpu

        forward_only = model is not None
        model = not_none_or_default(model, supplier=lambda: np.empty(3, dtype=np.int8))

        if self.model_type == self.MType_Scalar:
            n_cols = self.n_cells
        else:
            n_cols = self.n_cells * 3
            model = model.reshape(-1, 3)

        if forward_only:
            n_cols = 1

        n_rows = sum(receiver.n_rows for receiver in self.receivers)
        if is_cpu:
            ret = np.zeros((n_rows, n_cols), dtype=np.float64)
            xn = self.xn
            yn = self.yn
            zn = self.zn
            magnetization = self.magnetization
        else:
            # According to https://docs.taichi-lang.org/docs/ndarray, or
            # https://github.com/taichi-dev/taichi/blob/8fdf7a7d/docs/lang/articles/basic/ndarray.md?plain=1#L23C35-L23C35
            # ti.ndarray-s are initialized to 0 by default.
            ret = ti_ndarray(dtype=ti.f64, shape=(n_rows, n_cols))
            xn = ti_ndarray_from(self.xn)
            yn = ti_ndarray_from(self.yn)
            zn = ti_ndarray_from(self.zn)
            magnetization = ti_ndarray_from(self.magnetization)
        magnetization_dim = self.magnetization.ndim

        # TODO: 由于Taichi目前采用i32作为索引类型，数组元素总数不能超过int32的上限，否则行为未定义
        #  如果后续Taichi支持i64索引类型，则可以去掉这个限制
        ratio = n_rows * n_cols / ti_size_max
        assert ratio < 1, (
            f"Kernel matrix with more than `INT32_MAX` elements has exceeded limits of Taichi"
            f" ({ratio:.2f}x larger than `INT32_MAX`)."
            f"\nConsider set `store_sensitivities='forward_only'`"
            f" or `builder.store_sensitivities(False)` if using `SimulationBuilder`."
            f"\nSee also `https://github.com/taichi-dev/taichi/issues/8161`."
        )

        monitor = None
        if progress is not None:
            progress.reset(total=n_rows * n_cols)

            # 用于配合正演的分批计算
            check_start_row = 0  # 该批正演的起点行号
            ending_row = 0  # 该批正演的终点行号
            processed = 0  # 已经处理的元素个数

            # 假设1: 计算出的值在绝大多数情况下不为0
            # 假设2: 因为正演算子是按行优先顺序计算的，绝大多数情况下最后一列的计算会在之前所有元素同时或之后完成的
            # 因此可以通过监视线程读取返回值矩阵ret的最后一列来估算进度，将单次估测复杂度从O(RC)降低到O(R)
            # 另: 若正演算子按列优先，则可以监视最后一行，保证监控变量的内存邻近，但是这会导致正演计算效率下降，因此未采用
            monitor = ValueObserver(lambda: processed + np.count_nonzero(ret[check_start_row:ending_row, -1]) * n_cols,
                                    lambda dx: progress.update(dx),
                                    interval=0.5)

            if is_cpu:
                # GPU版本暂无法通过该方法实现，因此在下面通过sync手动每批更新一次
                # CPU版本则启动监视线程
                monitor.start()

        start_row = 0
        for receiver in self.receivers:
            receiver_locations = receiver.receiver_locations
            components = receiver.components

            # TODO: 由于Taichi目前采用i32作为索引类型，
            #  主循环使用的ti.ndrange也会受限于这个限制，
            #  因此需要分批处理，如果后续Taichi支持i64索引类型，则可以去掉这个限制
            # 注：该分块依据是ti.ndrange循环的下标，且只对forward_only有效，核矩阵的大小仍然会收到这个限制
            # TODO: 对于GPU版，可能需要基于GPU显存大小进行批次划分
            n_rx_rows = receiver.n_rows
            n_batches = int(np.ceil((n_rx_rows * self.n_cells) / ti_size_max))

            all_components = np.asarray(['tmi', 'bx', 'by', 'bz', 'bxx', 'bxy', 'bxz', 'byy', 'byz', 'bzz'])
            components_indices = {f'i{c}': -1 for c in all_components}
            for i, c in enumerate(receiver.components):
                components_indices[f'i{c}'] = i
            n_components = len(components)

            for rx_locs in np.array_split(receiver_locations, n_batches):
                n_receivers = rx_locs.shape[0]
                section_rows = n_receivers * n_components

                if monitor is not None:
                    # 计算进度判定行区间
                    # 区间的更新和已处理元素的更新需要同时进行，防止observer给出错误结果
                    ending_row += section_rows
                    check_start_row = ending_row - section_rows
                    processed = start_row * n_cols

                self._kernel_matrix_forward(
                    xn=xn, yn=yn, zn=zn,
                    receiver_locations=rx_locs,
                    model=model, forward_only=forward_only,
                    magnetization=magnetization,
                    magnetization_dim=magnetization_dim,
                    **components_indices, n_components=n_components,
                    start_row=start_row,
                    ret=ret)

                start_row += section_rows

                if monitor is not None:
                    # 同步正确进度，防止observer读取不及时，统一进度
                    # 并且在GPU版中用于更新进度
                    monitor.sync(start_row * n_cols)

        if monitor is not None:
            monitor.stop()
            progress.close()

        if not is_cpu:
            ret = ret.to_numpy()

        return ret

    @ensure_contiguous
    @ti_kernel
    def _kernel_matrix_forward(
            self,
            xn: ti.types.ndarray(), yn: ti.types.ndarray(), zn: ti.types.ndarray(),
            receiver_locations: ti.types.ndarray(),
            itmi: ti.template(), ibx: ti.template(), iby: ti.template(), ibz: ti.template(),
            ibxx: ti.template(), ibxy: ti.template(), ibxz: ti.template(),
            ibyy: ti.template(), ibyz: ti.template(),
            ibzz: ti.template(),
            n_components: ti.template(),
            magnetization: ti.types.ndarray(),
            magnetization_dim: ti.template(),
            model: ti.types.ndarray(),
            forward_only: ti.template(),
            start_row: ti.i64,
            ret: ti.types.ndarray(),
    ):
        """计算正演结果或核矩阵

        Parameters
        ----------
        xn, yn, zn
            网格x,y,z方向上下界坐标
        receiver_locations
            观测点网格
        itmi, ibx, iby, ibz, ibxx, ibxy, ibxz, ibyy, ibyz, ibzz
            每个分量在单个观测点数据中的索引，-1代表不需要计算该分量
        n_components
            需要计算的分量数
        magnetization
            磁化矩阵，array(n_cells, 3) or array(3,) or array()
        magnetization_dim
            磁化矩阵维度
        model
            模型参数，scalar模型下为array(n_cells,)，vector模型下为array(n_cells, 3)
        forward_only
            是否只计算正演结果，若为真则返回正演结果，否则返回核矩阵
        start_row
            起始行，指定输出矩阵在ret中的起点行

        Returns
        -------
        ret
            用于接收返回值，可以是numpy、torch或paddle的数组也可以是taichi的数组

        Notes
        -----
            非forward_only模式下，model参数不会参与计算
        """

        tol1 = 1e-10  # Tolerance 1 for numerical stability over nodes and edges
        tol2 = 1e-4  # Tolerance 2 for numerical stability over nodes and edges

        # number of observations
        n_obs = receiver_locations.shape[0]

        # base cell dimensions
        min_hx = self.base_cell_sizes[0]
        min_hy = self.base_cell_sizes[1]
        min_hz = self.base_cell_sizes[2]

        # projection for tmi, tmi = [bx, by, bz] * tmi_projection
        px = self.tmi_projection[0]
        py = self.tmi_projection[1]
        pz = self.tmi_projection[2]

        for iobs, icell in ti.ndrange(n_obs, self.n_cells):
            # comp. pos. differences for tne, bsw nodes. Adjust if location within
            # tolerance of a node or edge
            dz2 = zn[icell, 1] - receiver_locations[iobs, 2]
            if ti.abs(dz2) / min_hz < tol2:
                dz2 = tol2 * min_hz
            dz1 = zn[icell, 0] - receiver_locations[iobs, 2]
            if ti.abs(dz1) / min_hz < tol2:
                dz1 = tol2 * min_hz

            dy2 = yn[icell, 1] - receiver_locations[iobs, 1]
            if ti.abs(dy2) / min_hy < tol2:
                dy2 = tol2 * min_hy
            dy1 = yn[icell, 0] - receiver_locations[iobs, 1]
            if ti.abs(dy1) / min_hy < tol2:
                dy1 = tol2 * min_hy

            dx2 = xn[icell, 1] - receiver_locations[iobs, 0]
            if ti.abs(dx2) / min_hx < tol2:
                dx2 = tol2 * min_hx
            dx1 = xn[icell, 0] - receiver_locations[iobs, 0]
            if ti.abs(dx1) / min_hx < tol2:
                dx1 = tol2 * min_hx

            # comp. squared diff
            dx2dx2 = dx2 ** 2.0
            dx1dx1 = dx1 ** 2.0

            dy2dy2 = dy2 ** 2.0
            dy1dy1 = dy1 ** 2.0

            dz2dz2 = dz2 ** 2.0
            dz1dz1 = dz1 ** 2.0

            # 2D radius component squared of corner nodes
            R1 = dy2dy2 + dx2dx2
            R2 = dy2dy2 + dx1dx1
            R3 = dy1dy1 + dx2dx2
            R4 = dy1dy1 + dx1dx1

            # radius to each cell node
            r1 = ti.sqrt(dz2dz2 + R2)
            r2 = ti.sqrt(dz2dz2 + R1)
            r3 = ti.sqrt(dz1dz1 + R1)
            r4 = ti.sqrt(dz1dz1 + R2)
            r5 = ti.sqrt(dz2dz2 + R3)
            r6 = ti.sqrt(dz2dz2 + R4)
            r7 = ti.sqrt(dz1dz1 + R4)
            r8 = ti.sqrt(dz1dz1 + R3)

            # compactify argument calculations
            arg1_ = dx1 + dy2 + r1
            arg1 = dy2 + dz2 + r1
            arg2 = dx1 + dz2 + r1
            arg3 = dx1 + r1
            arg4 = dy2 + r1
            arg5 = dz2 + r1

            arg6_ = dx2 + dy2 + r2
            arg6 = dy2 + dz2 + r2
            arg7 = dx2 + dz2 + r2
            arg8 = dx2 + r2
            arg9 = dy2 + r2
            arg10 = dz2 + r2

            arg11_ = dx2 + dy2 + r3
            arg11 = dy2 + dz1 + r3
            arg12 = dx2 + dz1 + r3
            arg13 = dx2 + r3
            arg14 = dy2 + r3
            arg15 = dz1 + r3

            arg16_ = dx1 + dy2 + r4
            arg16 = dy2 + dz1 + r4
            arg17 = dx1 + dz1 + r4
            arg18 = dx1 + r4
            arg19 = dy2 + r4
            arg20 = dz1 + r4

            arg21_ = dx2 + dy1 + r5
            arg21 = dy1 + dz2 + r5
            arg22 = dx2 + dz2 + r5
            arg23 = dx2 + r5
            arg24 = dy1 + r5
            arg25 = dz2 + r5

            arg26_ = dx1 + dy1 + r6
            arg26 = dy1 + dz2 + r6
            arg27 = dx1 + dz2 + r6
            arg28 = dx1 + r6
            arg29 = dy1 + r6
            arg30 = dz2 + r6

            arg31_ = dx1 + dy1 + r7
            arg31 = dy1 + dz1 + r7
            arg32 = dx1 + dz1 + r7
            arg33 = dx1 + r7
            arg34 = dy1 + r7
            arg35 = dz1 + r7

            arg36_ = dx2 + dy1 + r8
            arg36 = dy1 + dz1 + r8
            arg37 = dx2 + dz1 + r8
            arg38 = dx2 + r8
            arg39 = dy1 + r8
            arg40 = dz1 + r8

            if ti.static(ibx >= 0) or ti.static(itmi >= 0):
                tx = (
                    -2 * ti.atan2(dx1, arg1 + tol1)
                    - -2 * ti.atan2(dx2, arg6 + tol1)
                    + -2 * ti.atan2(dx2, arg11 + tol1)
                    - -2 * ti.atan2(dx1, arg16 + tol1)
                    + -2 * ti.atan2(dx2, arg21 + tol1)
                    - -2 * ti.atan2(dx1, arg26 + tol1)
                    + -2 * ti.atan2(dx1, arg31 + tol1)
                    - -2 * ti.atan2(dx2, arg36 + tol1)
                ) / -4 / ti.math.pi

                ty = (
                    ti.log(arg5)
                    - ti.log(arg10)
                    + ti.log(arg15)
                    - ti.log(arg20)
                    + ti.log(arg25)
                    - ti.log(arg30)
                    + ti.log(arg35)
                    - ti.log(arg40)
                ) / -4 / ti.math.pi

                tz = (
                    ti.log(arg4) - ti.log(arg9)
                    + ti.log(arg14) - ti.log(arg19)
                    + ti.log(arg24) - ti.log(arg29)
                    + ti.log(arg34) - ti.log(arg39)
                ) / -4 / ti.math.pi

                if ti.static(ibx >= 0):
                    self.write_to_ret(
                        tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibx,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.NonAccumulative
                    )

                if ti.static(itmi >= 0):
                    self.write_to_ret(
                        tx=px * tx, ty=px * ty, tz=px * tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=itmi,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.Accumulative
                    )

            if ti.static(iby >= 0) or ti.static(itmi >= 0):
                tx = (
                    ti.log(arg5)
                    - ti.log(arg10)
                    + ti.log(arg15)
                    - ti.log(arg20)
                    + ti.log(arg25)
                    - ti.log(arg30)
                    + ti.log(arg35)
                    - ti.log(arg40)
                ) / -4 / ti.math.pi

                ty = (
                    -2 * ti.atan2(dy2, arg2 + tol1)
                    - -2 * ti.atan2(dy2, arg7 + tol1)
                    + -2 * ti.atan2(dy2, arg12 + tol1)
                    - -2 * ti.atan2(dy2, arg17 + tol1)
                    + -2 * ti.atan2(dy1, arg22 + tol1)
                    - -2 * ti.atan2(dy1, arg27 + tol1)
                    + -2 * ti.atan2(dy1, arg32 + tol1)
                    - -2 * ti.atan2(dy1, arg37 + tol1)
                ) / -4 / ti.math.pi

                tz = (
                    ti.log(arg3) - ti.log(arg8)
                    + ti.log(arg13) - ti.log(arg18)
                    + ti.log(arg23) - ti.log(arg28)
                    + ti.log(arg33) - ti.log(arg38)
                ) / -4 / ti.math.pi

                if ti.static(iby >= 0):
                    self.write_to_ret(
                        tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=iby,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.NonAccumulative
                    )

                if ti.static(itmi >= 0):
                    self.write_to_ret(
                        tx=py * tx, ty=py * ty, tz=py * tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=itmi,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.Accumulative
                    )

            if ti.static(ibz >= 0) or ti.static(itmi >= 0):
                tx = (
                    ti.log(arg4)
                    - ti.log(arg9)
                    + ti.log(arg14)
                    - ti.log(arg19)
                    + ti.log(arg24)
                    - ti.log(arg29)
                    + ti.log(arg34)
                    - ti.log(arg39)
                ) / -4 / ti.math.pi

                ty = (
                    ti.log(arg3) - ti.log(arg8)
                    + ti.log(arg13) - ti.log(arg18)
                    + ti.log(arg23) - ti.log(arg28)
                    + ti.log(arg33) - ti.log(arg38)
                ) / -4 / ti.math.pi

                tz = (
                    -2 * ti.atan2(dz2, arg1_ + tol1)
                    - -2 * ti.atan2(dz2, arg6_ + tol1)
                    + -2 * ti.atan2(dz1, arg11_ + tol1)
                    - -2 * ti.atan2(dz1, arg16_ + tol1)
                    + -2 * ti.atan2(dz2, arg21_ + tol1)
                    - -2 * ti.atan2(dz2, arg26_ + tol1)
                    + -2 * ti.atan2(dz1, arg31_ + tol1)
                    - -2 * ti.atan2(dz1, arg36_ + tol1)
                ) / -4 / ti.math.pi

                if ti.static(ibz >= 0):
                    self.write_to_ret(
                        tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibz,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.NonAccumulative
                    )

                if ti.static(itmi >= 0):
                    self.write_to_ret(
                        tx=pz * tx, ty=pz * ty, tz=pz * tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=itmi,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.Accumulative
                    )

            if ti.static(ibxx >= 0) or ti.static(ibzz >= 0):
                tx = 2 * (
                    ((dx1 ** 2 - r1 * arg1) / (r1 * arg1 ** 2 + dx1 ** 2 * r1))
                    - ((dx2 ** 2 - r2 * arg6) / (r2 * arg6 ** 2 + dx2 ** 2 * r2))
                    + ((dx2 ** 2 - r3 * arg11) / (r3 * arg11 ** 2 + dx2 ** 2 * r3))
                    - ((dx1 ** 2 - r4 * arg16) / (r4 * arg16 ** 2 + dx1 ** 2 * r4))
                    + ((dx2 ** 2 - r5 * arg21) / (r5 * arg21 ** 2 + dx2 ** 2 * r5))
                    - ((dx1 ** 2 - r6 * arg26) / (r6 * arg26 ** 2 + dx1 ** 2 * r6))
                    + ((dx1 ** 2 - r7 * arg31) / (r7 * arg31 ** 2 + dx1 ** 2 * r7))
                    - ((dx2 ** 2 - r8 * arg36) / (r8 * arg36 ** 2 + dx2 ** 2 * r8))
                ) / 4 / ti.math.pi

                ty = (
                    dx2 / (r5 * arg25)
                    - dx2 / (r2 * arg10)
                    + dx2 / (r3 * arg15)
                    - dx2 / (r8 * arg40)
                    + dx1 / (r1 * arg5)
                    - dx1 / (r6 * arg30)
                    + dx1 / (r7 * arg35)
                    - dx1 / (r4 * arg20)
                ) / 4 / ti.math.pi

                tz = (
                    dx1 / (r1 * arg4)
                    - dx2 / (r2 * arg9)
                    + dx2 / (r3 * arg14)
                    - dx1 / (r4 * arg19)
                    + dx2 / (r5 * arg24)
                    - dx1 / (r6 * arg29)
                    + dx1 / (r7 * arg34)
                    - dx2 / (r8 * arg39)
                ) / 4 / ti.math.pi

                if ti.static(ibxx >= 0):
                    self.write_to_ret(
                        tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibxx,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.NonAccumulative
                    )

                if ti.static(ibzz >= 0):
                    self.write_to_ret(
                        tx=-tx, ty=-ty, tz=-tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibzz,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.Accumulative
                    )

            if ti.static(ibyy >= 0) or ti.static(ibzz >= 0):
                tx = (
                    dy2 / (r3 * arg15)
                    - dy2 / (r2 * arg10)
                    + dy1 / (r5 * arg25)
                    - dy1 / (r8 * arg40)
                    + dy2 / (r1 * arg5)
                    - dy2 / (r4 * arg20)
                    + dy1 / (r7 * arg35)
                    - dy1 / (r6 * arg30)
                ) / 4 / ti.math.pi

                ty = 2 * (
                    ((dy2 ** 2 - r1 * arg2) / (r1 * arg2 ** 2 + dy2 ** 2 * r1))
                    - ((dy2 ** 2 - r2 * arg7) / (r2 * arg7 ** 2 + dy2 ** 2 * r2))
                    + ((dy2 ** 2 - r3 * arg12) / (r3 * arg12 ** 2 + dy2 ** 2 * r3))
                    - ((dy2 ** 2 - r4 * arg17) / (r4 * arg17 ** 2 + dy2 ** 2 * r4))
                    + ((dy1 ** 2 - r5 * arg22) / (r5 * arg22 ** 2 + dy1 ** 2 * r5))
                    - ((dy1 ** 2 - r6 * arg27) / (r6 * arg27 ** 2 + dy1 ** 2 * r6))
                    + ((dy1 ** 2 - r7 * arg32) / (r7 * arg32 ** 2 + dy1 ** 2 * r7))
                    - ((dy1 ** 2 - r8 * arg37) / (r8 * arg37 ** 2 + dy1 ** 2 * r8))
                ) / 4 / ti.math.pi

                tz = (
                    dy2 / (r1 * arg3)
                    - dy2 / (r2 * arg8)
                    + dy2 / (r3 * arg13)
                    - dy2 / (r4 * arg18)
                    + dy1 / (r5 * arg23)
                    - dy1 / (r6 * arg28)
                    + dy1 / (r7 * arg33)
                    - dy1 / (r8 * arg38)
                ) / 4 / ti.math.pi

                if ti.static(ibyy >= 0):
                    self.write_to_ret(
                        tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibyy,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.NonAccumulative
                    )

                if ti.static(ibzz >= 0):
                    self.write_to_ret(
                        tx=-tx, ty=-ty, tz=-tz, ret=ret, start_row=start_row,
                        icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibzz,
                        n_components=n_components,
                        model=model, forward_only=forward_only,
                        magnetization=magnetization, magnetization_dim=magnetization_dim,
                        accumulative=self.Accumulative
                    )

            if ti.static(ibxy >= 0):
                tx = 2 * (
                    ((dx1 * arg4) / (r1 * arg1 ** 2 + (dx1 ** 2) * r1))
                    - ((dx2 * arg9) / (r2 * arg6 ** 2 + (dx2 ** 2) * r2))
                    + ((dx2 * arg14) / (r3 * arg11 ** 2 + (dx2 ** 2) * r3))
                    - ((dx1 * arg19) / (r4 * arg16 ** 2 + (dx1 ** 2) * r4))
                    + ((dx2 * arg24) / (r5 * arg21 ** 2 + (dx2 ** 2) * r5))
                    - ((dx1 * arg29) / (r6 * arg26 ** 2 + (dx1 ** 2) * r6))
                    + ((dx1 * arg34) / (r7 * arg31 ** 2 + (dx1 ** 2) * r7))
                    - ((dx2 * arg39) / (r8 * arg36 ** 2 + (dx2 ** 2) * r8))
                ) / 4 / ti.math.pi

                ty = (
                    dy2 / (r1 * arg5)
                    - dy2 / (r2 * arg10)
                    + dy2 / (r3 * arg15)
                    - dy2 / (r4 * arg20)
                    + dy1 / (r5 * arg25)
                    - dy1 / (r6 * arg30)
                    + dy1 / (r7 * arg35)
                    - dy1 / (r8 * arg40)
                ) / 4 / ti.math.pi

                tz = (
                    1 / r1 - 1 / r2 + 1 / r3 - 1 / r4 + 1 / r5 - 1 / r6 + 1 / r7 - 1 / r8
                ) / 4 / ti.math.pi

                self.write_to_ret(
                    tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                    icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibxy,
                    n_components=n_components,
                    model=model, forward_only=forward_only,
                    magnetization=magnetization, magnetization_dim=magnetization_dim,
                    accumulative=self.NonAccumulative
                )

            if ti.static(ibxz >= 0):
                tx = 2 * (
                    ((dx1 * arg5) / (r1 * (arg1 ** 2) + (dx1 ** 2) * r1))
                    - ((dx2 * arg10) / (r2 * (arg6 ** 2) + (dx2 ** 2) * r2))
                    + ((dx2 * arg15) / (r3 * (arg11 ** 2) + (dx2 ** 2) * r3))
                    - ((dx1 * arg20) / (r4 * (arg16 ** 2) + (dx1 ** 2) * r4))
                    + ((dx2 * arg25) / (r5 * (arg21 ** 2) + (dx2 ** 2) * r5))
                    - ((dx1 * arg30) / (r6 * (arg26 ** 2) + (dx1 ** 2) * r6))
                    + ((dx1 * arg35) / (r7 * (arg31 ** 2) + (dx1 ** 2) * r7))
                    - ((dx2 * arg40) / (r8 * (arg36 ** 2) + (dx2 ** 2) * r8))
                ) / 4 / ti.math.pi

                ty = (
                    1 / r1 - 1 / r2 + 1 / r3 - 1 / r4 + 1 / r5 - 1 / r6 + 1 / r7 - 1 / r8
                ) / 4 / ti.math.pi

                tz = (
                    dz2 / (r1 * arg4)
                    - dz2 / (r2 * arg9)
                    + dz1 / (r3 * arg14)
                    - dz1 / (r4 * arg19)
                    + dz2 / (r5 * arg24)
                    - dz2 / (r6 * arg29)
                    + dz1 / (r7 * arg34)
                    - dz1 / (r8 * arg39)
                ) / 4 / ti.math.pi

                self.write_to_ret(
                    tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                    icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibxz,
                    n_components=n_components,
                    model=model, forward_only=forward_only,
                    magnetization=magnetization, magnetization_dim=magnetization_dim,
                    accumulative=self.NonAccumulative
                )

            if ti.static(ibyz >= 0):
                tx = (
                    1 / r3 - 1 / r2 + 1 / r5 - 1 / r8 + 1 / r1 - 1 / r4 + 1 / r7 - 1 / r6
                ) / 4 / ti.math.pi

                ty = 2 * (
                    ((dy2 * arg5) / (r1 * (arg2 ** 2) + (dy2 ** 2) * r1))
                    - ((dy2 * arg10) / (r2 * (arg7 ** 2) + (dy2 ** 2) * r2))
                    + ((dy2 * arg15) / (r3 * (arg12 ** 2) + (dy2 ** 2) * r3))
                    - ((dy2 * arg20) / (r4 * (arg17 ** 2) + (dy2 ** 2) * r4))
                    + ((dy1 * arg25) / (r5 * (arg22 ** 2) + (dy1 ** 2) * r5))
                    - ((dy1 * arg30) / (r6 * (arg27 ** 2) + (dy1 ** 2) * r6))
                    + ((dy1 * arg35) / (r7 * (arg32 ** 2) + (dy1 ** 2) * r7))
                    - ((dy1 * arg40) / (r8 * (arg37 ** 2) + (dy1 ** 2) * r8))
                ) / 4 / ti.math.pi

                tz = (
                    dz2 / (r1 * arg3)
                    - dz2 / (r2 * arg8)
                    + dz1 / (r3 * arg13)
                    - dz1 / (r4 * arg18)
                    + dz2 / (r5 * arg23)
                    - dz2 / (r6 * arg28)
                    + dz1 / (r7 * arg33)
                    - dz1 / (r8 * arg38)
                ) / 4 / ti.math.pi

                self.write_to_ret(
                    tx=tx, ty=ty, tz=tz, ret=ret, start_row=start_row,
                    icell=icell, iobs=iobs, n_obs=n_obs, icomponent=ibyz,
                    n_components=n_components,
                    model=model, forward_only=forward_only,
                    magnetization=magnetization, magnetization_dim=magnetization_dim,
                    accumulative=self.NonAccumulative
                )

    @ti_func
    def write_to_ret(
            self,
            tx, ty, tz,
            ret: ti.types.ndarray(),
            start_row,
            icell,
            iobs, n_obs,
            icomponent: ti.template(),
            model: ti.types.ndarray(),
            magnetization: ti.types.ndarray(),
            magnetization_dim: ti.template(),
            n_components: ti.template(),
            forward_only: ti.template(),
            accumulative: ti.template(),
    ):
        """将正演结果写入输出矩阵

        Parameters
        ----------
        tx, ty, tz
            三方向系数
        ret
            输出矩阵
        start_row
            输出矩阵的起点行
        magnetization
            磁化矩阵，array(n_cells, 3) or array(3,) or array()
        magnetization_dim
            磁化矩阵维度
        model
            模型数组，如果为scalar模型则应为array(n_cells,)否则应为array(n_cells, 3)
        icell
            网格索引
        iobs
            观测点索引
        icomponent
            分量索引
        n_obs
            观测点数量
        n_components
            待求解分量数量
        forward_only
            是否只计算正演（直接乘以模型）
        accumulative
            是否累加（如果为False则会覆盖原有值），适用于forward_only或tmi的计算
        """
        i = self.in_col_index(iobs, icomponent, n_components, n_obs) + start_row

        # 根据磁化矩阵的维度选择计算方式
        if ti.static(magnetization_dim == 2):
            tx *= magnetization[icell, 0]
            ty *= magnetization[icell, 1]
            tz *= magnetization[icell, 2]
        elif ti.static(magnetization_dim == 1):
            tx *= magnetization[0]
            ty *= magnetization[1]
            tz *= magnetization[2]
        else:
            tx *= magnetization[None]
            ty *= magnetization[None]
            tz *= magnetization[None]

        if ti.static(self.model_type == 0):  # MType_Vector
            if ti.static(not forward_only):
                jx = self.in_row_index(icell, 0)
                jy = self.in_row_index(icell, 1)
                jz = self.in_row_index(icell, 2)
                if ti.static(accumulative == 0):  # NonAccumulative
                    ret[i, jx] = tx
                    ret[i, jy] = ty
                    ret[i, jz] = tz
                else:
                    ret[i, jx] += tx
                    ret[i, jy] += ty
                    ret[i, jz] += tz
            else:
                ret[i, 0] += model[icell, 0] * tx + model[icell, 1] * ty + model[icell, 2] * tz
        else:
            t = tx + ty + tz

            if ti.static(not forward_only):
                if ti.static(accumulative == 0):  # NonAccumulative
                    ret[i, icell] = t
                else:
                    ret[i, icell] += t
            else:
                ret[i, 0] += t * model[icell]

    @ti_func
    def in_col_index(self, iobs, icomponent, n_components, n_obs):
        return ti_index(iobs, icomponent, self.col_stype, n_components, n_obs)

    @ti_func
    def in_row_index(self, icell, bi):
        return ti_index(icell, bi, self.row_stype, 3, self.n_cells)



import os

import numpy as np
import torch

import psutil

from metalpy.mexin import Mixin
from metalpy.scab.progressed import Progress
from metalpy.scab.utils.sparse import scipy2torch

PROFILE_MEMORY = False


class TorchedImpl(Mixin):
    def __init__(self, this):
        super().__init__(this)

        if PROFILE_MEMORY:
            from memory_profiler import profile
            self.evaluator = profile(self._evaluate_integral)
        else:
            self.evaluator = self._evaluate_integral

    @staticmethod
    def estimate_memory_cost(this, size=None):
        size = this.survey.receiver_locations.shape[0] if size is None else size
        return 309.1 * 1024 * 1024 / 10 / 50250 * size * this.modelMap.shape[0]

    def linear_operator(this, self):
        def create_tensor(x, dtype=None):
            # return torch.tensor(x, dtype=dtype, device=self.get_torch_device())
            if dtype is None:
                return torch.from_numpy(x).to(self.get_torch_device())
            else:
                return torch.as_tensor(x, dtype=dtype, device=self.get_torch_device())

        # infos = [torch.cuda.get_device_properties(i) for i in range(ng)]
        # 获取显存大小来实施计算分块

        self.nC = self.modelMap.shape[0]  # 网格数
        nD = self.survey.nD  # 数据量，观测点数 * 目标量数
        nObs = self.survey.receiver_locations.shape[0]  # self.nD  # 观测点数

        if self.store_sensitivities == "disk":
            sens_name = self.sensitivity_path + "sensitivity.npy"
            if os.path.exists(sens_name):
                # do not pull array completely into ram, just need to check the size
                kernel = np.load(sens_name, mmap_mode="r")
                if kernel.shape == (nD, self.nC):
                    print(f"Found sensitivity file at {sens_name} with expected shape")
                    kernel = np.asarray(kernel)
                    return kernel
        # Single threaded
        mem = psutil.virtual_memory().free
        mem -= 100 * 1024 * 1024  # 100MB
        chunk_size = self.estimate_batch_size(mem)
        n_chunks = int(np.ceil(nObs / chunk_size))

        receiver_lists = np.array_split(self.survey.receiver_locations, n_chunks)
        active_components = {
            k: np.array_split(v, n_chunks) for k, v in self.survey.components.items()
        }

        kernels = []
        for i, receivers in enumerate(receiver_lists):
            components = {k: v[i] for k, v in active_components.items()}
            kernel = this.evaluator(create_tensor(receivers),
                                    {k: create_tensor(v) for k, v in
                                    components.items()}, )
            kernels.append(kernel)

            self.mixins.get_mixin_or_sentinel(Progress).update(len(receivers))

        kernel = torch.vstack(kernels)
        model = create_tensor(self.model)

        if self.store_sensitivities == "forward_only":
            kernel = torch.einsum('ocg,g->oc', kernel, model)

        if self.store_sensitivities == "disk":
            print(f"writing sensitivity to {sens_name}")
            os.makedirs(self.sensitivity_path, exist_ok=True)
            np.save(sens_name, kernel)

        return kernel.detach().numpy()


    def _evaluate_integral(_, self, receiver_location, components):
        """
        Load in the active nodes of a tensor mesh and computes the magnetic
        forward relation between a cuboid and a given observation
        location outside the Earth [obsx, obsy, obsz]

        INPUT:
        receiver_location:  [obsx, obsy, obsz] nC x 3 Array

        components: list[str]
            List of magnetic components chosen from:
            'bx', 'by', 'bz', 'bxx', 'bxy', 'bxz', 'byy', 'byz', 'bzz'

        OUTPUT:
        Tx = [Txx Txy Txz]
        Ty = [Tyx Tyy Tyz]
        Tz = [Tzx Tzy Tzz]
        """

        def create_tensor(x, dtype=None):
            # return torch.tensor(x, dtype=dtype, device=self.get_torch_device())
            if dtype is None:
                return torch.from_numpy(x).to(self.get_torch_device())
            else:
                return torch.as_tensor(x, dtype=dtype, device=self.get_torch_device())

        def create_zeros(shape, dtype=None):
            return torch.zeros(shape, dtype=torch.float64 if dtype is None else dtype, device=self.get_torch_device())

        # TODO: This should probably be converted to C
        tol1 = 1e-10  # Tolerance 1 for numerical stability over nodes and edges
        tol2 = 1e-4  # Tolerance 2 for numerical stability over nodes and edges

        receiver_location.requires_grad = False
        # number of cells in mesh
        nC = self.Xn.shape[0]
        batch_size = receiver_location.shape[0]

        # TODO: 用component mask控制具体点上要计算的分量
        rows = {}

        # base cell dimensions
        min_hx, min_hy, min_hz = (
            self.mesh.hx.min(),
            self.mesh.hy.min(),
            self.mesh.hz.min(),
        )

        # comp. pos. differences for tne, bsw nodes. Adjust if location within
        # tolerance of a node or edge
        # 数据打包，如果没有性能问题就算了
        # boundaries = np.c_[self.Xn, self.Yn, self.Zn]
        # boundaries = create_tensor(boundaries)
        #
        # d2 = boundaries[:, ::2]

        # self.M是CSR，转置后变成CSC
        # 使用torch的CSR/CSC格式存在以下问题
        # 1. CSR计算结果有问题（核心问题）
        # 2. CSR只能左乘（不过COO也只能左乘）
        # 3. CSR的操作无法求导
        # TODO: 但是CSR理论计算效率高于COO，如果能解决上述问题，那么这里可以考虑不用转换为COO
        M = scipy2torch(self.M.T, device=self.get_torch_device()).to_sparse_coo()

        def mul_by_m(vd, key):
            # truth = torch.as_tensor(vd[key].detach().numpy() * self.M)
            mult = M @ vd[key].T
            mult = mult.T
            vd[key] = mult

        Zn = create_tensor(self.Zn)
        Zs = receiver_location[:, 2][:, None, None].expand(-1, nC, -1)
        dz = Zn - Zs
        del Zn, Zs
        dz[torch.abs(dz) < tol2 * min_hz] = tol2 * min_hz
        dzdz = dz ** 2
        dz1, dz2 = dz[:, :, 0], dz[:, :, 1]
        dz1dz1, dz2dz2 = dzdz[:, :, 0], dzdz[:, :, 1]

        Yn = create_tensor(self.Yn)
        Ys = receiver_location[:, 1][:, None, None].expand(-1, nC, -1)
        dy = Yn - Ys
        del Yn, Ys
        dy[torch.abs(dy) < tol2 * min_hy] = tol2 * min_hy
        dydy = dy ** 2
        dy1, dy2 = dy[:, :, 0], dy[:, :, 1]
        dy1dy1, dy2dy2 = dydy[:, :, 0], dydy[:, :, 1]

        Xn = create_tensor(self.Xn)
        Xs = receiver_location[:, 0][:, None, None].expand(-1, nC, -1)
        dx = Xn - Xs
        del Xn, Xs
        dx[torch.abs(dx) < tol2 * min_hx] = tol2 * min_hx
        dxdx = dx ** 2
        dx1, dx2 = dx[:, :, 0], dx[:, :, 1]
        dx1dx1, dx2dx2 = dxdx[:, :, 0], dxdx[:, :, 1]

        # 2D radius component squared of corner nodes
        R1 = dy2dy2 + dx2dx2
        R2 = dy2dy2 + dx1dx1
        R3 = dy1dy1 + dx2dx2
        R4 = dy1dy1 + dx1dx1

        # radius to each cell node
        r1 = torch.sqrt(dz2dz2 + R2)
        r2 = torch.sqrt(dz2dz2 + R1)
        r3 = torch.sqrt(dz1dz1 + R1)
        r4 = torch.sqrt(dz1dz1 + R2)
        r5 = torch.sqrt(dz2dz2 + R3)
        r6 = torch.sqrt(dz2dz2 + R4)
        r7 = torch.sqrt(dz1dz1 + R4)
        r8 = torch.sqrt(dz1dz1 + R3)

        # compactify argument calculations
        arg1_ = dx1 + dy2 + r1
        arg1 = dy2 + dz2 + r1
        arg2 = dx1 + dz2 + r1
        arg3 = dx1 + r1
        arg4 = dy2 + r1
        arg5 = dz2 + r1

        arg6_ = dx2 + dy2 + r2
        arg6 = dy2 + dz2 + r2
        arg7 = dx2 + dz2 + r2
        arg8 = dx2 + r2
        arg9 = dy2 + r2
        arg10 = dz2 + r2

        arg11_ = dx2 + dy2 + r3
        arg11 = dy2 + dz1 + r3
        arg12 = dx2 + dz1 + r3
        arg13 = dx2 + r3
        arg14 = dy2 + r3
        arg15 = dz1 + r3

        arg16_ = dx1 + dy2 + r4
        arg16 = dy2 + dz1 + r4
        arg17 = dx1 + dz1 + r4
        arg18 = dx1 + r4
        arg19 = dy2 + r4
        arg20 = dz1 + r4

        arg21_ = dx2 + dy1 + r5
        arg21 = dy1 + dz2 + r5
        arg22 = dx2 + dz2 + r5
        arg23 = dx2 + r5
        arg24 = dy1 + r5
        arg25 = dz2 + r5

        arg26_ = dx1 + dy1 + r6
        arg26 = dy1 + dz2 + r6
        arg27 = dx1 + dz2 + r6
        arg28 = dx1 + r6
        arg29 = dy1 + r6
        arg30 = dz2 + r6

        arg31_ = dx1 + dy1 + r7
        arg31 = dy1 + dz1 + r7
        arg32 = dx1 + dz1 + r7
        arg33 = dx1 + r7
        arg34 = dy1 + r7
        arg35 = dz1 + r7

        arg36_ = dx2 + dy1 + r8
        arg36 = dy1 + dz1 + r8
        arg37 = dx2 + dz1 + r8
        arg38 = dx2 + r8
        arg39 = dy1 + r8
        arg40 = dz1 + r8

        if ("bxx" in components) or ("bzz" in components):
            rows["bxx"] = create_zeros((batch_size, 3 * nC))

            rows["bxx"][:, 0:nC] = 2 * (
                    ((dx1 ** 2 - r1 * arg1) / (r1 * arg1 ** 2 + dx1 ** 2 * r1))
                    - ((dx2 ** 2 - r2 * arg6) / (r2 * arg6 ** 2 + dx2 ** 2 * r2))
                    + ((dx2 ** 2 - r3 * arg11) / (r3 * arg11 ** 2 + dx2 ** 2 * r3))
                    - ((dx1 ** 2 - r4 * arg16) / (r4 * arg16 ** 2 + dx1 ** 2 * r4))
                    + ((dx2 ** 2 - r5 * arg21) / (r5 * arg21 ** 2 + dx2 ** 2 * r5))
                    - ((dx1 ** 2 - r6 * arg26) / (r6 * arg26 ** 2 + dx1 ** 2 * r6))
                    + ((dx1 ** 2 - r7 * arg31) / (r7 * arg31 ** 2 + dx1 ** 2 * r7))
                    - ((dx2 ** 2 - r8 * arg36) / (r8 * arg36 ** 2 + dx2 ** 2 * r8))
            )

            rows["bxx"][:, nC: 2 * nC] = (
                    dx2 / (r5 * arg25)
                    - dx2 / (r2 * arg10)
                    + dx2 / (r3 * arg15)
                    - dx2 / (r8 * arg40)
                    + dx1 / (r1 * arg5)
                    - dx1 / (r6 * arg30)
                    + dx1 / (r7 * arg35)
                    - dx1 / (r4 * arg20)
            )

            rows["bxx"][:, 2 * nC:] = (
                    dx1 / (r1 * arg4)
                    - dx2 / (r2 * arg9)
                    + dx2 / (r3 * arg14)
                    - dx1 / (r4 * arg19)
                    + dx2 / (r5 * arg24)
                    - dx1 / (r6 * arg29)
                    + dx1 / (r7 * arg34)
                    - dx2 / (r8 * arg39)
            )

            rows["bxx"] /= 4 * torch.pi
            mul_by_m(rows, "bxx")

        if ("byy" in components) or ("bzz" in components):
            rows["byy"] = create_zeros((batch_size, 3 * nC))

            rows["byy"][:, 0:nC] = (
                    dy2 / (r3 * arg15)
                    - dy2 / (r2 * arg10)
                    + dy1 / (r5 * arg25)
                    - dy1 / (r8 * arg40)
                    + dy2 / (r1 * arg5)
                    - dy2 / (r4 * arg20)
                    + dy1 / (r7 * arg35)
                    - dy1 / (r6 * arg30)
            )
            rows["byy"][:, nC: 2 * nC] = 2 * (
                    ((dy2 ** 2 - r1 * arg2) / (r1 * arg2 ** 2 + dy2 ** 2 * r1))
                    - ((dy2 ** 2 - r2 * arg7) / (r2 * arg7 ** 2 + dy2 ** 2 * r2))
                    + ((dy2 ** 2 - r3 * arg12) / (r3 * arg12 ** 2 + dy2 ** 2 * r3))
                    - ((dy2 ** 2 - r4 * arg17) / (r4 * arg17 ** 2 + dy2 ** 2 * r4))
                    + ((dy1 ** 2 - r5 * arg22) / (r5 * arg22 ** 2 + dy1 ** 2 * r5))
                    - ((dy1 ** 2 - r6 * arg27) / (r6 * arg27 ** 2 + dy1 ** 2 * r6))
                    + ((dy1 ** 2 - r7 * arg32) / (r7 * arg32 ** 2 + dy1 ** 2 * r7))
                    - ((dy1 ** 2 - r8 * arg37) / (r8 * arg37 ** 2 + dy1 ** 2 * r8))
            )
            rows["byy"][:, 2 * nC:] = (
                    dy2 / (r1 * arg3)
                    - dy2 / (r2 * arg8)
                    + dy2 / (r3 * arg13)
                    - dy2 / (r4 * arg18)
                    + dy1 / (r5 * arg23)
                    - dy1 / (r6 * arg28)
                    + dy1 / (r7 * arg33)
                    - dy1 / (r8 * arg38)
            )

            rows["byy"] /= 4 * torch.pi
            mul_by_m(rows, "byy")

        if "bzz" in components:
            rows["bzz"] = -rows["bxx"] - rows["byy"]

        if "bxy" in components:
            rows["bxy"] = create_zeros((batch_size, 3 * nC))

            rows["bxy"][:, 0:nC] = 2 * (
                    ((dx1 * arg4) / (r1 * arg1 ** 2 + (dx1 ** 2) * r1))
                    - ((dx2 * arg9) / (r2 * arg6 ** 2 + (dx2 ** 2) * r2))
                    + ((dx2 * arg14) / (r3 * arg11 ** 2 + (dx2 ** 2) * r3))
                    - ((dx1 * arg19) / (r4 * arg16 ** 2 + (dx1 ** 2) * r4))
                    + ((dx2 * arg24) / (r5 * arg21 ** 2 + (dx2 ** 2) * r5))
                    - ((dx1 * arg29) / (r6 * arg26 ** 2 + (dx1 ** 2) * r6))
                    + ((dx1 * arg34) / (r7 * arg31 ** 2 + (dx1 ** 2) * r7))
                    - ((dx2 * arg39) / (r8 * arg36 ** 2 + (dx2 ** 2) * r8))
            )
            rows["bxy"][:, nC: 2 * nC] = (
                    dy2 / (r1 * arg5)
                    - dy2 / (r2 * arg10)
                    + dy2 / (r3 * arg15)
                    - dy2 / (r4 * arg20)
                    + dy1 / (r5 * arg25)
                    - dy1 / (r6 * arg30)
                    + dy1 / (r7 * arg35)
                    - dy1 / (r8 * arg40)
            )
            rows["bxy"][:, 2 * nC:] = (
                    1 / r1 - 1 / r2 + 1 / r3 - 1 / r4 + 1 / r5 - 1 / r6 + 1 / r7 - 1 / r8
            )

            rows["bxy"] /= 4 * torch.pi
            mul_by_m(rows, "bxy")

        if "bxz" in components:
            rows["bxz"] = create_zeros((batch_size, 3 * nC))

            rows["bxz"][:, 0:nC] = 2 * (
                    ((dx1 * arg5) / (r1 * (arg1 ** 2) + (dx1 ** 2) * r1))
                    - ((dx2 * arg10) / (r2 * (arg6 ** 2) + (dx2 ** 2) * r2))
                    + ((dx2 * arg15) / (r3 * (arg11 ** 2) + (dx2 ** 2) * r3))
                    - ((dx1 * arg20) / (r4 * (arg16 ** 2) + (dx1 ** 2) * r4))
                    + ((dx2 * arg25) / (r5 * (arg21 ** 2) + (dx2 ** 2) * r5))
                    - ((dx1 * arg30) / (r6 * (arg26 ** 2) + (dx1 ** 2) * r6))
                    + ((dx1 * arg35) / (r7 * (arg31 ** 2) + (dx1 ** 2) * r7))
                    - ((dx2 * arg40) / (r8 * (arg36 ** 2) + (dx2 ** 2) * r8))
            )
            rows["bxz"][:, nC: 2 * nC] = (
                    1 / r1 - 1 / r2 + 1 / r3 - 1 / r4 + 1 / r5 - 1 / r6 + 1 / r7 - 1 / r8
            )
            rows["bxz"][:, 2 * nC:] = (
                    dz2 / (r1 * arg4)
                    - dz2 / (r2 * arg9)
                    + dz1 / (r3 * arg14)
                    - dz1 / (r4 * arg19)
                    + dz2 / (r5 * arg24)
                    - dz2 / (r6 * arg29)
                    + dz1 / (r7 * arg34)
                    - dz1 / (r8 * arg39)
            )

            rows["bxz"] /= 4 * torch.pi
            mul_by_m(rows, "bxz")

        if "byz" in components:
            rows["byz"] = create_zeros((batch_size, 3 * nC))

            rows["byz"][:, 0:nC] = (
                    1 / r3 - 1 / r2 + 1 / r5 - 1 / r8 + 1 / r1 - 1 / r4 + 1 / r7 - 1 / r6
            )
            rows["byz"][:, nC: 2 * nC] = 2 * (
                    (dy2 * arg5 / r1 * (arg2 ** 2) + (dy2 ** 2) * r1)
                    - ((dy2 * arg10 / r2 * (arg7 ** 2) + (dy2 ** 2) * r2))
                    + ((dy2 * arg15 / r3 * (arg12 ** 2) + (dy2 ** 2) * r3))
                    - ((dy2 * arg20 / r4 * (arg17 ** 2) + (dy2 ** 2) * r4))
                    + ((dy1 * arg25 / r5 * (arg22 ** 2) + (dy1 ** 2) * r5))
                    - ((dy1 * arg30 / r6 * (arg27 ** 2) + (dy1 ** 2) * r6))
                    + ((dy1 * arg35 / r7 * (arg32 ** 2) + (dy1 ** 2) * r7))
                    - ((dy1 * arg40 / r8 * (arg37 ** 2) + (dy1 ** 2) * r8))
            )
            rows["byz"][:, 2 * nC:] = (
                    dz2 / (r1 * arg3)
                    - dz2 / (r2 * arg8)
                    + dz1 / (r3 * arg13)
                    - dz1 / (r4 * arg18)
                    + dz2 / (r5 * arg23)
                    - dz2 / (r6 * arg28)
                    + dz1 / (r7 * arg33)
                    - dz1 / (r8 * arg38)
            )

            rows["byz"] /= 4 * torch.pi
            mul_by_m(rows, "byz")

        if ("bx" in components) or ("tmi" in components):
            rows["bx"] = create_zeros((batch_size, 3 * nC))

            rows["bx"][:, 0:nC] = (
                    (-2 * torch.arctan2(dx1, arg1 + tol1))
                    - (-2 * torch.arctan2(dx2, arg6 + tol1))
                    + (-2 * torch.arctan2(dx2, arg11 + tol1))
                    - (-2 * torch.arctan2(dx1, arg16 + tol1))
                    + (-2 * torch.arctan2(dx2, arg21 + tol1))
                    - (-2 * torch.arctan2(dx1, arg26 + tol1))
                    + (-2 * torch.arctan2(dx1, arg31 + tol1))
                    - (-2 * torch.arctan2(dx2, arg36 + tol1))
            )
            rows["bx"][:, nC: 2 * nC] = (
                    torch.log(arg5)
                    - torch.log(arg10)
                    + torch.log(arg15)
                    - torch.log(arg20)
                    + torch.log(arg25)
                    - torch.log(arg30)
                    + torch.log(arg35)
                    - torch.log(arg40)
            )
            rows["bx"][:, 2 * nC:] = (
                    (torch.log(arg4) - torch.log(arg9))
                    + (torch.log(arg14) - torch.log(arg19))
                    + (torch.log(arg24) - torch.log(arg29))
                    + (torch.log(arg34) - torch.log(arg39))
            )
            rows["bx"] /= -4 * torch.pi
            mul_by_m(rows, "bx")

        if ("by" in components) or ("tmi" in components):
            rows["by"] = create_zeros((batch_size, 3 * nC))

            rows["by"][:, 0:nC] = (
                    torch.log(arg5)
                    - torch.log(arg10)
                    + torch.log(arg15)
                    - torch.log(arg20)
                    + torch.log(arg25)
                    - torch.log(arg30)
                    + torch.log(arg35)
                    - torch.log(arg40)
            )
            rows["by"][:, nC: 2 * nC] = (
                    (-2 * torch.arctan2(dy2, arg2 + tol1))
                    - (-2 * torch.arctan2(dy2, arg7 + tol1))
                    + (-2 * torch.arctan2(dy2, arg12 + tol1))
                    - (-2 * torch.arctan2(dy2, arg17 + tol1))
                    + (-2 * torch.arctan2(dy1, arg22 + tol1))
                    - (-2 * torch.arctan2(dy1, arg27 + tol1))
                    + (-2 * torch.arctan2(dy1, arg32 + tol1))
                    - (-2 * torch.arctan2(dy1, arg37 + tol1))
            )
            rows["by"][:, 2 * nC:] = (
                    (torch.log(arg3) - torch.log(arg8))
                    + (torch.log(arg13) - torch.log(arg18))
                    + (torch.log(arg23) - torch.log(arg28))
                    + (torch.log(arg33) - torch.log(arg38))
            )

            rows["by"] /= -4 * torch.pi
            mul_by_m(rows, "by")

        if ("bz" in components) or ("tmi" in components):
            rows["bz"] = create_zeros((batch_size, 3 * nC))

            rows["bz"][:, 0:nC] = (
                    torch.log(arg4)
                    - torch.log(arg9)
                    + torch.log(arg14)
                    - torch.log(arg19)
                    + torch.log(arg24)
                    - torch.log(arg29)
                    + torch.log(arg34)
                    - torch.log(arg39)
            )
            rows["bz"][:, nC: 2 * nC] = (
                    (torch.log(arg3) - torch.log(arg8))
                    + (torch.log(arg13) - torch.log(arg18))
                    + (torch.log(arg23) - torch.log(arg28))
                    + (torch.log(arg33) - torch.log(arg38))
            )
            rows["bz"][:, 2 * nC:] = (
                    (-2 * torch.arctan2(dz2, arg1_ + tol1))
                    - (-2 * torch.arctan2(dz2, arg6_ + tol1))
                    + (-2 * torch.arctan2(dz1, arg11_ + tol1))
                    - (-2 * torch.arctan2(dz1, arg16_ + tol1))
                    + (-2 * torch.arctan2(dz2, arg21_ + tol1))
                    - (-2 * torch.arctan2(dz2, arg26_ + tol1))
                    + (-2 * torch.arctan2(dz1, arg31_ + tol1))
                    - (-2 * torch.arctan2(dz1, arg36_ + tol1))
            )
            rows["bz"] /= -4 * torch.pi
            mul_by_m(rows, "bz")

        if "tmi" in components:
            # truth = self.evaluate_integral(receiver_location=receiver_location[0], components=list(components.keys()))
            mat = torch.concat([rows[c][:, :, None] for c in ['bx', 'by', 'bz']], dim=2)
            proj = create_tensor(self.tmi_projection)

            rows["tmi"] = torch.einsum('ij,bkj->bk', proj, mat)

        return torch.concat([rows[component][:, None, :] for component in components], dim=1)



from __future__ import annotations

from typing import Any, Union

from metalpy.utils.dhash import dhash
from metalpy.utils.model import pv_ufunc_assign, DataAssociation
from .mix_modes import MixMode, Mixer, dhashable_mixer
from .shapes import Shape3D


class Object:
    DEFAULT_KEY = '__DEFAULT__'

    def __init__(self,
                 shape: Shape3D,
                 models: Union[dict[str, Any], Any] | None = None,
                 mix_mode: Mixer = MixMode.Override):
        """代表一个三维几何体图层

        Parameters
        ----------
        shape
            三维几何体
        models
            三维几何体的参数
        mix_mode
            混合模式，必须为MixMode枚举或自定义的混合函数，函数的参数见 Object.mix
        """
        self.shape = shape
        if models is None:
            models = True
        if not isinstance(models, dict):
            models = {Object.DEFAULT_KEY: models}
        self.models = models

        self.mix_mode = mix_mode

    def to_polydata(self):
        shape = self.shape
        poly = shape.to_polydata()

        if poly is not None:
            for k, v in self.models.items():
                pv_ufunc_assign(poly, DataAssociation.Cell, k, v, inplace=True)

        return poly

    @property
    def mixer(self):
        return MixMode.dispatch(self.mix_mode)

    @property
    def shape(self) -> Shape3D:
        return self._shape

    @shape.setter
    def shape(self, val: Shape3D):
        self._shape = val

    @property
    def models(self) -> dict[str, Any]:
        return self._models

    @models.setter
    def models(self, val: dict[str, Any]):
        self._models = val

    @property
    def model(self) -> Any:
        return self.models[Object.DEFAULT_KEY]

    @model.setter
    def model(self, val: Any):
        self.models[Object.DEFAULT_KEY] = val

    @property
    def n_tasks(self):
        return self.shape.n_tasks

    @property
    def progress_manually(self):
        return self.shape.progress_manually

    def items(self):
        for k, v in self.models.items():
            yield k, v

    def mix(self, prev_layer, current_layer):
        """执行模型混合操作

        Parameters
        ----------
        prev_layer
            之前所有模型结果和当前模型的重合部分的值数组
        current_layer
            当前模型重合部分的值数据

        Returns
        -------
        ret
            混合操作后的结果
        """
        return self.mixer(prev_layer, current_layer)

    def __getitem__(self, item):
        return self.models[item]

    def __dhash__(self):
        return dhash(self.shape, dhashable_mixer(self.mix_mode), self.models)
